"""CF representation of metadata according to Sentinel-1 Product Specification.

See: S1-RS-MDA-52-7441, DI-MPC-PB, MPC-0240, 3/7, 27/02/2020
https://sentinel.esa.int/documents/247904/1877131/Sentinel-1-Product-Specification
"""

import xarray as xr

from . import __version__

GROUP_ATTRIBUTES = {
    "orbit": {
        "title": "Orbit information used by the IPF during processing",
        "comment": (
            "The dataset contains a sets of orbit state vectors that are updated along azimuth."
            " The values represent the interpolated values used by the IPF"
            " and are derived from the sub-commutated ancillary data from the ISPs"
            " or from an input auxiliary orbit file"
        ),
    },
    "attitude": {
        "title": "Attitude information used by the IPF during processing",
        "comment": (
            "The dataset contains a sets of attitude data records that are updated along azimuth."
            " The values represent the interpolated values used by the IPF"
            " and are derived from the sub-commutated ancillary data from the ISPs"
            " or from an input auxiliary orbit file"
        ),
    },
    "gcp": {
        "title": "Geolocation grid",
        "comment": (
            "The dataset contains geolocation grid point entries for each line/pixel"
            " combination based on a configured resolution."
            " The list contains an entry for each update made along azimuth"
        ),
    },
    "calibration": {
        "title": "Calibration coefficients",
        "comment": (
            "The dataset contains calibration information and the beta nought, sigma nought,"
            " gamma and digital number (DN) Look-up Tables (LUT) that can be used for"
            " absolute product calibration"
        ),
    },
}

VARIABLE_ATTRIBUTES = {
    "line": {"units": "1", "long_name": "product line number"},
    "pixel": {"units": "1", "long_name": "product pixel number"},
    "azimuth_time": {"long_name": "zero-Doppler azimuth time", "standard_name": "time"},
    # NOTE: `slant_range_time` is not expressed as `np.timedelta64[ns]` in order to keep enough
    #   accuracy for interferometric processing, i.e. c * 1ns / 2 ~= 15cm.
    "slant_range_time": {"units": "s", "long_name": "slant range time / two-way delay"},
    "ground_range": {"units": "m", "long_name": "ground range"},
    "axis": {"units": "1", "long_name": "coordinate index"},
    "degree": {"units": "1", "long_name": "polynomial degree"},
    "latitude": {"units": "degrees_north", "long_name": "geodetic latitude"},
    "longitude": {"units": "degrees_east", "long_name": "geodetic longitude"},
    "height": {"units": "m", "long_name": "height above sea level"},
    "incidenceAngle": {"units": "°", "long_name": "incidence angle"},
    "elevationAngle": {"units": "°", "long_name": "elevation angle"},
    "q0": {"units": "1", "long_name": "Q0 attitude quaternion"},
    "q1": {"units": "1", "long_name": "Q1 attitude quaternion"},
    "q2": {"units": "1", "long_name": "Q2 attitude quaternion"},
    "q3": {"units": "1", "long_name": "Q3 attitude quaternion"},
    "roll": {"units": "°", "long_name": "platform roll"},
    "pitch": {"units": "°", "long_name": "platform pitch"},
    "yaw": {"units": "°", "long_name": "platform yaw"},
    "wx": {"units": "° s-1", "long_name": "X component of angular velocity vector"},
    "wy": {"units": "° s-1", "long_name": "Y component of angular velocity vector"},
    "wz": {"units": "° s-1", "long_name": "Z component of angular velocity vector"},
    "position": {"units": "m", "long_name": "ECEF position"},
    "velocity": {"units": "m s-1", "long_name": "ECEF velocity"},
    "sigmaNought": {"units": "m m-1", "long_name": "sigma nought calibration LUT"},
    "betaNought": {"units": "m m-1", "long_name": "beta nought calibration LUT"},
    "gamma": {"units": "m m-1", "long_name": "gamma calibration LUT"},
    "dn": {"units": "1", "long_name": "original digital number calibration LUT"},
    "noiseRangeLut": {"units": "1", "long_name": "range thermal noise correction LUT"},
    "noiseAzimuthLut": {
        "units": "1",
        "long_name": "azimuth thermal noise correction LUT",
    },
    "gr0": {
        "units": "m",
        "long_name": "ground range origin for slant range calculation",
    },
    "grsrCoefficients": {
        "units": "1",
        "long_name": "polynomial to convert from ground range to slant range",
    },
    "sr0": {
        "units": "m",
        "long_name": "slant range origin for ground range calculation",
    },
    "srgrCoefficients": {
        "units": "1",
        "long_name": "polynomial to convert from slant range to ground range",
    },
    "t0": {
        "units": "s",
        "long_name": "Two-way slant range time origin",
    },
    "data_dc_polynomial": {
        "units": "1",
        "long_name": "Doppler centroid estimated from data",
    },
    "geometry_dc_polynomial": {
        "units": "1",
        "long_name": "Doppler centroid estimated from geometry",
    },
    "data_dc_rms_error": {
        "units": "1",
        "long_name": "RMS error of the Doppler centroid estimate",
    },
    "data_dc_rms_error_above_threshold": {
        "units": "1",
        "long_name": "RMS error below the acceptable threshold for Doppler centroid estimated from the data",
    },
    "azimuth_fm_rate_polynomial": {
        "units": "1",
        "long_name": "Azimuth FM rate coefficient array",
    },
    "measurement": {"units": "1", "long_name": "digital number"},
}


def update_attributes(ds: xr.Dataset, group: str = "") -> xr.Dataset:
    # NOTE: keep the version in sync with the capabilities of CF compliance checkers
    ds.attrs["Conventions"] = "CF-1.11"
    ds.attrs.update(GROUP_ATTRIBUTES.get(group, {}))
    ds.attrs["history"] = f"created by xarray_sentinel-{__version__}"
    for var in ds.variables:
        attrs = VARIABLE_ATTRIBUTES.get(str(var), {})
        ds.variables[var].attrs.update(attrs)
    return ds



import pathlib
from typing import Any

import pytest

from xarray_sentinel import reformat

pytest.importorskip("netCDF4")

DATA_FOLDER = pathlib.Path(__file__).parent / "data"


def test_to_group_netcdf(tmpdir: Any) -> None:
    product_path = (
        DATA_FOLDER
        / "S1B_IW_GRDH_1SDV_20210401T052623_20210401T052648_026269_032297_ECC8.SAFE"
    )
    tmp_path = str(tmpdir.join("tmp.nc"))
    groups = {"IW/VV/gcp": "IW/VV/gcp", "IW/VH/attitude": "IW/VH/attitude"}

    reformat.to_group_netcdf(product_path, tmp_path, groups, engine="netcdf4")

    reformat.to_group_netcdf(product_path, tmp_path, engine="netcdf4")



import pathlib

import pytest
import xarray as xr

from xarray_sentinel import esa_safe

DATA_FOLDER = pathlib.Path(__file__).parent / "data"


COMMON_ATTRIBUTES = {
    "family_name": "SENTINEL-1",
    "number": "B",
    "mode": "IW",
    "swaths": ["IW1", "IW2", "IW3"],
    "orbit_number": 26269,
    "relative_orbit_number": 168,
    "pass": "DESCENDING",
    "ascending_node_time": "2021-04-01T04:49:55.637823",
    "transmitter_receiver_polarisations": ["VV", "VH"],
    "product_type": "SLC",
}


SENTINEL1_SLC_PRODUCTS = [
    (
        DATA_FOLDER
        / "S1B_IW_SLC__1SDV_20210401T052622_20210401T052650_026269_032297_EFA4.SAFE",
        "IW1/VV",
    ),
    (
        DATA_FOLDER
        / "S1A_EW_SLC__1SDH_20210403T122536_20210403T122630_037286_046484_8152.SAFE",
        "EW1/HH",
    ),
    (
        DATA_FOLDER
        / "S1A_S3_SLC__1SDV_20210401T152855_20210401T152914_037258_04638E_6001.SAFE",
        "S3/VH",
    ),
]

SENTINEL1_GRD_PRODUCTS = [
    (
        DATA_FOLDER
        / "S1B_IW_GRDH_1SDV_20210401T052623_20210401T052648_026269_032297_ECC8.SAFE",
        "IW/VV",
    ),
]
SENTINEL1_PRODUCTS = SENTINEL1_SLC_PRODUCTS + SENTINEL1_GRD_PRODUCTS


def test_open_dataset_root() -> None:
    product_path = (
        DATA_FOLDER
        / "S1B_IW_SLC__1SDV_20210401T052622_20210401T052650_026269_032297_EFA4.SAFE"
    )
    res = xr.open_dataset(product_path, engine="sentinel-1")

    assert isinstance(res, xr.Dataset)
    for attr_name in COMMON_ATTRIBUTES:
        assert attr_name in res.attrs
        assert res.attrs[attr_name] == COMMON_ATTRIBUTES[attr_name]

    res = xr.open_dataset(product_path)

    assert isinstance(res, xr.Dataset)

    product_path = product_path / "manifest.safe"

    res = xr.open_dataset(product_path, engine="sentinel-1")

    assert isinstance(res, xr.Dataset)

    res = xr.open_dataset(product_path)

    assert isinstance(res, xr.Dataset)

    with pytest.raises((ValueError, FileNotFoundError)):
        xr.open_dataset("")


@pytest.mark.parametrize("product_path,swath_pol", SENTINEL1_SLC_PRODUCTS)
def test_open_dataset_polarisation_slc(
    product_path: esa_safe.PathType,
    swath_pol: str,
) -> None:
    res = xr.open_dataset(product_path, engine="sentinel-1", group=swath_pol)

    assert isinstance(res, xr.Dataset)
    assert set(res.sizes) == {"line", "pixel"} or set(res.sizes) == {
        "azimuth_time",
        "slant_range_time",
    }
    assert set(res.coords) == {"azimuth_time", "slant_range_time", "line", "pixel"}


@pytest.mark.parametrize("product_path,swath_pol", SENTINEL1_GRD_PRODUCTS)
def test_open_dataset_polarisation_grd(
    product_path: esa_safe.PathType,
    swath_pol: str,
) -> None:
    res = xr.open_dataset(product_path, engine="sentinel-1", group=swath_pol)

    assert isinstance(res, xr.Dataset)
    assert set(res.sizes) == {"line", "pixel"} or set(res.sizes) == {
        "azimuth_time",
        "ground_range",
    }
    assert set(res.coords) == {"azimuth_time", "ground_range", "line", "pixel"}


@pytest.mark.parametrize("product_path,swath_pol", SENTINEL1_PRODUCTS)
def test_open_dataset_orbit(
    product_path: esa_safe.PathType,
    swath_pol: str,
) -> None:
    res = xr.open_dataset(product_path, engine="sentinel-1", group=f"{swath_pol}/orbit")

    assert isinstance(res, xr.Dataset)
    assert set(res.sizes) == {"axis", "azimuth_time"}
    assert set(res.variables) == {"azimuth_time", "axis", "velocity", "position"}


@pytest.mark.parametrize("product_path,swath_pol", SENTINEL1_PRODUCTS)
def test_open_dataset_attitude(
    product_path: esa_safe.PathType,
    swath_pol: str,
) -> None:
    res = xr.open_dataset(
        product_path, engine="sentinel-1", group=f"{swath_pol}/attitude"
    )

    assert isinstance(res, xr.Dataset)
    assert set(res.sizes) == {"azimuth_time"}
    expected = {
        "azimuth_time",
        "roll",
        "pitch",
        "yaw",
        "q0",
        "q1",
        "q2",
        "q3",
        "wx",
        "wy",
        "wz",
    }
    assert set(res.variables) == expected


@pytest.mark.parametrize("product_path,swath_pol", SENTINEL1_PRODUCTS)
def test_open_dataset_gcp(
    product_path: esa_safe.PathType,
    swath_pol: str,
) -> None:
    res = xr.open_dataset(product_path, engine="sentinel-1", group=f"{swath_pol}/gcp")

    assert isinstance(res, xr.Dataset)
    assert set(res.sizes) == {"azimuth_time", "slant_range_time"}


@pytest.mark.parametrize("product_path,swath_pol", SENTINEL1_PRODUCTS)
def test_open_dataset_dc_estimate(
    product_path: esa_safe.PathType,
    swath_pol: str,
) -> None:
    res = xr.open_dataset(
        product_path, engine="sentinel-1", group=f"{swath_pol}/dc_estimate"
    )

    assert isinstance(res, xr.Dataset)
    assert set(res.sizes) == {"azimuth_time", "degree"}


def test_open_pol_dataset() -> None:
    product_path = (
        DATA_FOLDER
        / "S1B_IW_SLC__1SDV_20210401T052622_20210401T052650_026269_032297_EFA4.SAFE"
    )
    expected_variables = {
        "measurement",
        "line",
        "pixel",
        "slant_range_time",
        "azimuth_time",
    }
    res = xr.open_dataset(product_path, engine="sentinel-1", group="IW1/VV")

    assert isinstance(res, xr.Dataset)
    for attr_name in COMMON_ATTRIBUTES:
        assert attr_name in res.attrs
        assert res.attrs[attr_name] == COMMON_ATTRIBUTES[attr_name]

    assert set(res.sizes) == {"line", "pixel"}
    assert set(res.variables) == expected_variables


def test_burst_id_attribute() -> None:
    product_path = (
        DATA_FOLDER
        / "S1A_IW_SLC__1SDH_20220414T102209_20220414T102236_042768_051AA4_E677.SAFE"
    )

    res = xr.open_dataset(product_path, engine="sentinel-1", group="IW1/HH")
    assert "burst_ids" in res.attrs
    assert len(res.attrs["burst_ids"]) == res.attrs["number_of_bursts"]

    product_path = (
        DATA_FOLDER
        / "S1B_IW_SLC__1SDV_20210401T052622_20210401T052650_026269_032297_EFA4.SAFE"
    )
    res = xr.open_dataset(product_path, engine="sentinel-1", group="IW1/VV")
    assert "burst_ids" not in res.attrs


def test_open_calibration_dataset() -> None:
    annotation_path = (
        DATA_FOLDER
        / "S1B_IW_SLC__1SDV_20210401T052622_20210401T052650_026269_032297_EFA4.SAFE"
    )
    res = xr.open_dataset(
        annotation_path, engine="sentinel-1", group="IW1/VV/calibration"
    )

    assert isinstance(res, xr.Dataset)
    assert set(res.sizes) == {"line", "pixel"}



import pathlib

import numpy as np
import pytest
import shapely.geometry
import shapely.wkt
import xarray as xr
from stac_validator import stac_validator

from xarray_sentinel import esa_safe, sentinel1

DATA_FOLDER = pathlib.Path(__file__).parent / "data"

SLC_IW_V340 = (
    DATA_FOLDER
    / "S1A_IW_SLC__1SDH_20220414T102209_20220414T102236_042768_051AA4_E677.SAFE"
)

SLC_IW = (
    DATA_FOLDER
    / "S1B_IW_SLC__1SDV_20210401T052622_20210401T052650_026269_032297_EFA4.SAFE"
)
SLC_IW1_VV_annotation = (
    SLC_IW
    / "annotation"
    / "s1b-iw1-slc-vv-20210401t052624-20210401t052649-026269-032297-004.xml"
)
SLC_IW1_VV_calibration = (
    SLC_IW
    / "annotation"
    / "calibration"
    / "calibration-s1b-iw1-slc-vv-20210401t052624-20210401t052649-026269-032297-004.xml"
)
SLC_IW1_VV_noise = (
    SLC_IW
    / "annotation"
    / "calibration"
    / "noise-s1b-iw1-slc-vv-20210401t052624-20210401t052649-026269-032297-004.xml"
)
SLC_IW1_VV_measurement = (
    SLC_IW
    / "measurement"
    / "s1b-iw1-slc-vv-20210401t052624-20210401t052649-026269-032297-004.tiff"
)
SLC_S3 = (
    DATA_FOLDER
    / "S1A_S3_SLC__1SDV_20210401T152855_20210401T152914_037258_04638E_6001.SAFE"
)
SLC_S3_VH_annotation = (
    SLC_S3
    / "annotation"
    / "s1a-s3-slc-vh-20210401t152855-20210401t152914-037258-04638e-001.xml"
)
SLC_S3_VH_measurement = (
    SLC_S3
    / "measurement"
    / "s1a-s3-slc-vh-20210401t152855-20210401t152914-037258-04638e-001.tiff"
)
GRD_IW = (
    DATA_FOLDER
    / "S1B_IW_GRDH_1SDV_20210401T052623_20210401T052648_026269_032297_ECC8.SAFE"
)
GRD_IW_VV_annotation = (
    GRD_IW
    / "annotation"
    / "s1b-iw-grd-vv-20210401t052623-20210401t052648-026269-032297-001.xml"
)


def test_get_fs_path() -> None:
    fs, path = sentinel1.get_fs_path(str(SLC_IW))

    assert path == str((SLC_IW / "manifest.safe").absolute())

    fs, path = sentinel1.get_fs_path(SLC_IW, fs)

    assert path == str((SLC_IW / "manifest.safe"))

    with pytest.raises(TypeError):
        sentinel1.get_fs_path("*", fs=fs, storage_options={})

    with pytest.raises(ValueError):
        sentinel1.get_fs_path("non-existent-path/*")

    with pytest.raises(ValueError):
        sentinel1.get_fs_path(DATA_FOLDER / "*")


def test_normalise_group() -> None:
    assert sentinel1.normalise_group(None) == ("", None)
    assert sentinel1.normalise_group("/") == ("", None)
    assert sentinel1.normalise_group("IW1") == ("IW1", None)
    assert sentinel1.normalise_group("/IW1") == ("IW1", None)
    assert sentinel1.normalise_group("/IW1/VH/0") == ("IW1/VH", 0)
    assert sentinel1.normalise_group("/IW1/VH/orbit") == ("IW1/VH/orbit", None)


def test_open_calibration_dataset() -> None:
    res = sentinel1.open_calibration_dataset(SLC_IW1_VV_calibration)

    assert isinstance(res, xr.Dataset)
    assert set(res.coords) == {"line", "pixel"}


def test_open_noise_range_dataset() -> None:
    res = sentinel1.open_noise_range_dataset(SLC_IW1_VV_noise)

    assert isinstance(res, xr.Dataset)
    assert set(res.coords) == {"line", "pixel"}


def test_open_noise_azimuth_dataset() -> None:
    res = sentinel1.open_noise_azimuth_dataset(SLC_IW1_VV_noise)

    assert isinstance(res, xr.Dataset)
    assert set(res.coords) == {"line"}


def test_open_coordinate_conversion_dataset() -> None:
    res = sentinel1.open_coordinate_conversion_dataset(GRD_IW_VV_annotation)

    assert isinstance(res, xr.Dataset)
    assert set(res.coords) == {"azimuth_time", "degree"}

    with pytest.raises(TypeError):
        sentinel1.open_coordinate_conversion_dataset(SLC_IW1_VV_annotation)


def test_open_gcp_dataset() -> None:
    expected_geospatial_bounds = (
        "POLYGON((11.26870151724317 47.24053130234206,"
        "10.876144717121 45.73265733767158,12.04397933341514 45.57910451206848,"
        "12.42647347821595 47.09200435560957,11.26870151724317 47.24053130234206))"
    )
    expected_polygon = shapely.wkt.loads(expected_geospatial_bounds)
    expected_geospatial_bbox = (
        10.876144717121,
        45.57910451206848,
        12.42647347821595,
        47.24053130234206,
    )

    res = sentinel1.open_gcp_dataset(SLC_IW1_VV_annotation)

    assert isinstance(res, xr.Dataset)
    assert set(res.coords) == {"line", "pixel", "azimuth_time", "slant_range_time"}
    assert isinstance(res.attrs["geospatial_bounds"], str)
    assert shapely.wkt.loads(res.attrs["geospatial_bounds"]).is_valid
    assert shapely.wkt.loads(res.attrs["geospatial_bounds"]).equals(expected_polygon)
    assert res.attrs["geospatial_bounds"] == expected_geospatial_bounds
    geospatial_bbox = (
        res.attrs["geospatial_lon_min"],
        res.attrs["geospatial_lat_min"],
        res.attrs["geospatial_lon_max"],
        res.attrs["geospatial_lat_max"],
    )
    assert np.allclose(geospatial_bbox, expected_geospatial_bbox)


def test_get_footprint_linestring() -> None:
    gcp_ds = sentinel1.open_gcp_dataset(SLC_IW1_VV_annotation)
    expected_linestring = [
        (11.26870151724317, 47.24053130234206),
        (10.876144717121, 45.73265733767158),
        (12.04397933341514, 45.57910451206848),
        (12.42647347821595, 47.09200435560957),
        (11.26870151724317, 47.24053130234206),
    ]

    res = sentinel1.get_footprint_linestring(
        gcp_ds.azimuth_time, gcp_ds.slant_range_time, gcp_ds
    )

    polygon = shapely.geometry.Polygon(res)
    assert polygon.is_valid
    assert shapely.geometry.polygon.orient(polygon, 1) == polygon
    assert res == expected_linestring

    res = sentinel1.get_footprint_linestring(
        gcp_ds.azimuth_time[::-1], gcp_ds.slant_range_time, gcp_ds
    )

    polygon = shapely.geometry.Polygon(res)
    assert polygon.is_valid
    assert shapely.geometry.polygon.orient(polygon, 1) == polygon


def test_open_attitude_dataset() -> None:
    res = sentinel1.open_attitude_dataset(SLC_IW1_VV_annotation)

    assert isinstance(res, xr.Dataset)
    assert set(res.coords) == {"azimuth_time"}


def test_open_orbit_dataset() -> None:
    res = sentinel1.open_orbit_dataset(SLC_IW1_VV_annotation)

    assert isinstance(res, xr.Dataset)
    assert set(res.coords) == {"axis", "azimuth_time"}


def test_open_dc_estimate_dataset() -> None:
    res = sentinel1.open_dc_estimate_dataset(SLC_IW1_VV_annotation)

    assert isinstance(res, xr.Dataset)
    assert set(res.coords) == {"degree", "azimuth_time"}


def test_open_azimuth_fm_rate_dataset() -> None:
    res = sentinel1.open_azimuth_fm_rate_dataset(SLC_IW1_VV_annotation)

    assert isinstance(res, xr.Dataset)
    assert set(res.coords) == {"degree", "azimuth_time"}


def test_open_pol_dataset_iw() -> None:
    res = sentinel1.open_pol_dataset(SLC_IW1_VV_measurement, SLC_IW1_VV_annotation)

    assert isinstance(res, xr.Dataset)
    assert set(res.sizes) == {"line", "pixel"}
    assert set(res.coords) == {"slant_range_time", "azimuth_time", "line", "pixel"}

    first_line = np.datetime64(res.attrs["product_first_line_utc_time"])
    last_line = np.datetime64(res.attrs["product_last_line_utc_time"])
    assert res.azimuth_time[0] == first_line
    assert res.azimuth_time[-1] == last_line

    expected_geospatial_bounds = (
        "POLYGON((11.26870151724317 47.24053130234206,"
        "10.876144717121 45.73265733767158,12.04397933341514 45.57910451206848,"
        "12.42647347821595 47.09200435560957,11.26870151724317 47.24053130234206))"
    )
    gcp_ds = sentinel1.open_gcp_dataset(SLC_IW1_VV_annotation)

    res = sentinel1.open_pol_dataset(
        SLC_IW1_VV_measurement, SLC_IW1_VV_annotation, gcp=gcp_ds
    )

    assert res.attrs["geospatial_bounds"] == expected_geospatial_bounds


def test_open_pol_dataset_sm() -> None:
    res = sentinel1.open_pol_dataset(SLC_S3_VH_measurement, SLC_S3_VH_annotation)

    assert isinstance(res, xr.Dataset)
    assert set(res.sizes) == {"slant_range_time", "azimuth_time"}
    assert set(res.coords) == {"slant_range_time", "azimuth_time", "line", "pixel"}

    first_line = np.datetime64(res.attrs["product_first_line_utc_time"])
    last_line = np.datetime64(res.attrs["product_last_line_utc_time"])
    assert res.azimuth_time[0] == first_line
    assert res.azimuth_time[-1] == last_line


def test_find_avalable_groups() -> None:
    _, product_files = esa_safe.parse_manifest_sentinel1(SLC_S3 / "manifest.safe")
    expected_groups = {
        "S3",
        "S3/VV",
        "S3/VV/attitude",
        "S3/VV/gcp",
        "S3/VV/orbit",
        "S3/VV/dc_estimate",
        "S3/VV/azimuth_fm_rate",
        "S3/VV/calibration",
        "S3/VV/noise_range",
        "S3/VV/noise_azimuth",
        "S3/VH",
        "S3/VH/attitude",
        "S3/VH/gcp",
        "S3/VH/orbit",
        "S3/VH/dc_estimate",
        "S3/VH/azimuth_fm_rate",
        "S3/VH/calibration",
        "S3/VH/noise_range",
        "S3/VH/noise_azimuth",
    }

    res = sentinel1.find_available_groups(product_files, str(SLC_IW), "SLC")

    assert set(res) == expected_groups

    res = sentinel1.find_available_groups(
        product_files, str(SLC_IW), "SLC", check_files_exist=True
    )
    assert res == {}


def test_open_sentinel1_dataset() -> None:
    expected_groups = {
        "IW1",
        "IW1/VV",
        "IW1/VV/gcp",
        "IW1/VV/attitude",
        "IW1/VV/orbit",
        "IW1/VV/calibration",
        "IW1/VV/noise_range",
        "IW1/VV/noise_azimuth",
    }

    res = sentinel1.open_sentinel1_dataset(SLC_IW)

    assert isinstance(res, xr.Dataset)
    assert len(res.data_vars) == 0
    assert set(res.attrs["subgroups"]) >= expected_groups

    res = sentinel1.open_sentinel1_dataset(SLC_IW, group="IW1")

    assert isinstance(res, xr.Dataset)
    assert len(res.data_vars) == 0

    res = sentinel1.open_sentinel1_dataset(SLC_IW, group="IW1/VV/orbit")

    assert isinstance(res, xr.Dataset)
    assert res.sizes == {"axis": 3, "azimuth_time": 17}

    with pytest.raises(ValueError):
        sentinel1.open_sentinel1_dataset(SLC_IW, group="IW1/VV/non-existent")


def test_open_dataset_virtual_groups() -> None:
    res = sentinel1.open_sentinel1_dataset(SLC_IW, group="IW1/VV/0")

    assert isinstance(res, xr.Dataset)
    assert len(res.data_vars) == 1
    assert res.attrs["burst_index"] == 0


def test_crop_burst_dataset() -> None:
    swath_ds = sentinel1.open_sentinel1_dataset(SLC_IW_V340, group="IW1/HH")

    res1 = sentinel1.crop_burst_dataset(swath_ds, 8)

    assert set(res1.sizes) == {"azimuth_time", "slant_range_time"}
    assert res1.sizes["azimuth_time"] == swath_ds.attrs["lines_per_burst"]

    res2 = sentinel1.crop_burst_dataset(swath_ds, azimuth_anx_time=2210)

    assert res2.equals(res1)

    res3 = sentinel1.crop_burst_dataset(
        swath_ds, azimuth_anx_time=2213, use_center=True
    )

    assert res3.equals(res1)

    res4 = sentinel1.crop_burst_dataset(swath_ds, burst_id=365923)

    assert res4.equals(res1)

    with pytest.raises(TypeError):
        sentinel1.crop_burst_dataset(swath_ds)

    with pytest.raises(TypeError):
        sentinel1.crop_burst_dataset(swath_ds, burst_index=8, azimuth_anx_time=2213)

    with pytest.raises(TypeError):
        sentinel1.crop_burst_dataset(swath_ds, burst_index=8, burst_id=365923)

    with pytest.raises(TypeError):
        sentinel1.crop_burst_dataset(swath_ds, azimuth_anx_time=2213, burst_id=365923)

    with pytest.raises(IndexError):
        sentinel1.crop_burst_dataset(swath_ds, burst_index=-1)

    with pytest.raises(KeyError):
        sentinel1.crop_burst_dataset(swath_ds, burst_id=1)

    swath_ds = sentinel1.open_sentinel1_dataset(SLC_IW, group="IW1/VH")

    with pytest.raises(TypeError):
        sentinel1.crop_burst_dataset(swath_ds, burst_id=1)


def test_crop_burst_dataset_gcp() -> None:
    expected_geospatial_bounds = (
        "POLYGON((11.060741069073366 46.41270558692681,"
        "11.010711557932021 46.228050403919134,12.175715782110206 46.076023778486665,"
        "12.209682909257705 46.26327129178887,11.060741069073366 46.41270558692681))"
    )
    expected_polygon = shapely.wkt.loads(expected_geospatial_bounds)
    expected_geospatial_bbox = (
        11.010711557932021,
        46.076023778486665,
        12.209682909257705,
        46.41270558692681,
    )

    swath_ds = sentinel1.open_sentinel1_dataset(SLC_IW, group="IW1/VV")
    gcp_ds = sentinel1.open_sentinel1_dataset(SLC_IW, group="IW1/VV/gcp")

    res = sentinel1.crop_burst_dataset(swath_ds, burst_index=5, gcp=gcp_ds)

    assert set(res.sizes) == {"azimuth_time", "slant_range_time"}
    assert res.sizes["azimuth_time"] == swath_ds.attrs["lines_per_burst"]
    assert isinstance(res.attrs["geospatial_bounds"], str)
    assert shapely.wkt.loads(res.attrs["geospatial_bounds"]).is_valid
    assert shapely.wkt.loads(res.attrs["geospatial_bounds"]).equals(expected_polygon)
    assert res.attrs["geospatial_bounds"] == expected_geospatial_bounds
    geospatial_bbox = (
        res.attrs["geospatial_lon_min"],
        res.attrs["geospatial_lat_min"],
        res.attrs["geospatial_lon_max"],
        res.attrs["geospatial_lat_max"],
    )
    assert np.allclose(geospatial_bbox, expected_geospatial_bbox)


def test_mosaic_slc_iw() -> None:
    ds = sentinel1.open_sentinel1_dataset(SLC_IW_V340, group="IW1/HH")

    res = sentinel1.mosaic_slc_iw(ds)

    assert isinstance(res, xr.Dataset)

    res = sentinel1.mosaic_slc_iw(ds.measurement)

    assert isinstance(res, xr.DataArray)


def test_calibrate_amplitude() -> None:
    swath_ds = sentinel1.open_sentinel1_dataset(SLC_IW, group="IW1/VH")
    burst_ds = sentinel1.crop_burst_dataset(swath_ds, burst_index=8)
    cal_ds = sentinel1.open_sentinel1_dataset(SLC_IW, group="IW1/VH/calibration")

    res = sentinel1.calibrate_amplitude(burst_ds.measurement, cal_ds["betaNought"])

    assert set(res.sizes) == {"azimuth_time", "slant_range_time"}
    assert np.issubdtype(res.dtype, np.complex64)

    res = sentinel1.calibrate_amplitude(burst_ds.measurement, cal_ds["gamma"])

    assert set(res.sizes) == {"azimuth_time", "slant_range_time"}
    assert np.issubdtype(res.dtype, np.complex64)


def test_calibrate_intensity() -> None:
    swath_ds = sentinel1.open_sentinel1_dataset(SLC_IW, group="IW1/VH")
    burst_ds = sentinel1.crop_burst_dataset(swath_ds, burst_index=8)
    cal_ds = sentinel1.open_sentinel1_dataset(SLC_IW, group="IW1/VH/calibration")

    res = sentinel1.calibrate_intensity(burst_ds.measurement, cal_ds["betaNought"])

    assert set(res.sizes) == {"azimuth_time", "slant_range_time"}
    assert np.issubdtype(res.dtype, np.float32)

    cal_ds["betaNought"].attrs.pop("long_name")

    res = sentinel1.calibrate_intensity(
        burst_ds.measurement, cal_ds["betaNought"], as_db=True
    )

    assert set(res.sizes) == {"azimuth_time", "slant_range_time"}
    assert np.issubdtype(res.dtype, np.float32)

    res = sentinel1.calibrate_intensity(
        burst_ds.measurement,
        cal_ds["betaNought"],
        as_db=True,
        min_db=None,
    )

    assert np.issubdtype(res.dtype, np.float32)


def test_slant_range_time_to_ground_range() -> None:
    swath_ds = sentinel1.open_sentinel1_dataset(SLC_IW, group="IW1/VV")
    swath = swath_ds.measurement[:1000, :1000]
    cc_ds = sentinel1.open_sentinel1_dataset(
        GRD_IW, group="IW/VV/coordinate_conversion"
    )

    res = sentinel1.slant_range_time_to_ground_range(
        swath.azimuth_time, swath.slant_range_time, cc_ds
    )

    assert isinstance(res, xr.DataArray)


def test_ground_range_to_slant_range_time() -> None:
    swath_ds = sentinel1.open_sentinel1_dataset(GRD_IW, group="IW/VV")
    swath = swath_ds.measurement[:1000, :1000]
    cc_ds = sentinel1.open_sentinel1_dataset(
        GRD_IW, group="IW/VV/coordinate_conversion"
    )

    res = sentinel1.ground_range_to_slant_range_time(
        swath.azimuth_time, swath.ground_range, cc_ds
    )

    assert isinstance(res, xr.DataArray)


def test_do_override_product_files() -> None:
    template = "{dirname}/{prefix}{swath}-{polarization}{ext}"
    _, product_files = esa_safe.parse_manifest_sentinel1(SLC_S3 / "manifest.safe")

    res = sentinel1.do_override_product_files(template, product_files)

    assert "./annotation/s3-vv.xml" in res


@pytest.mark.parametrize(
    "safe,annotation",
    [
        [SLC_S3, SLC_S3_VH_annotation],
        [SLC_IW, SLC_IW1_VV_annotation],
        [GRD_IW, GRD_IW_VV_annotation],
    ],
)
def test_make_sentinel1_stac_item(safe: pathlib.Path, annotation: pathlib.Path) -> None:
    item = sentinel1.make_sentinel1_stac_item(
        "test", safe / "manifest.safe", annotation
    )

    validator = stac_validator.StacValidate()
    # HACK: the geometry appears to be correct, but it doesn't validate
    result = validator.validate_dict(item | {"geometry": None})

    assert result, validator.message



import pathlib

import pytest
import xarray as xr

dask = pytest.importorskip("dask")

DATA_FOLDER = pathlib.Path(__file__).parent / "data"


def test_open_pol_dataset_iw_preferred_chunks() -> None:
    product_path = (
        DATA_FOLDER
        / "S1B_IW_SLC__1SDV_20210401T052622_20210401T052650_026269_032297_EFA4.SAFE"
    )
    res = xr.open_dataset(product_path, engine="sentinel-1", group="IW1/VV", chunks={})

    assert isinstance(res, xr.Dataset)
    assert len(res.dims) == 2
    line_index = res.measurement.dims.index("line")
    assert res.measurement.chunks[line_index][0] == 1

    res = xr.open_dataset(
        product_path,
        engine="sentinel-1",
        group="IW1/VV",
        chunks={},
        rasterio_chunks={"y": 64},
    )

    assert isinstance(res, xr.Dataset)
    assert len(res.dims) == 2
    line_index = res.measurement.dims.index("line")
    assert res.measurement.chunks[line_index][0] == 64


def test_open_pol_dataset_sm_preferred_chunks() -> None:
    product_path = (
        DATA_FOLDER
        / "S1A_S3_SLC__1SDV_20210401T152855_20210401T152914_037258_04638E_6001.SAFE"
    )
    res = xr.open_dataset(product_path, engine="sentinel-1", group="S3/VH", chunks={})

    assert isinstance(res, xr.Dataset)
    assert len(res.dims) == 2
    azimuth_time_index = res.measurement.dims.index("azimuth_time")
    assert res.measurement.chunks[azimuth_time_index][0] == 1

    res = xr.open_dataset(
        product_path,
        engine="sentinel-1",
        group="S3/VH",
        chunks={},
        rasterio_chunks={"y": 64},
    )

    assert isinstance(res, xr.Dataset)
    assert len(res.dims) == 2
    azimuth_time_index = res.measurement.dims.index("azimuth_time")
    assert res.measurement.chunks[azimuth_time_index][0] == 64



import pathlib
from typing import Any, Dict
from xml.etree import ElementTree

import pytest
import xmlschema

from xarray_sentinel import esa_safe

DATA_FOLDER = pathlib.Path(__file__).parent / "data"

SENTINEL1_ATTRIBUTES = {
    "S1B_IW_SLC__1SDV_20210401T052622_20210401T052650_026269_032297_EFA4": {
        "ascending_node_time": "2021-04-01T04:49:55.637823",
        "family_name": "SENTINEL-1",
        "mode": "IW",
        "number": "B",
        "orbit_number": 26269,
        "pass": "DESCENDING",
        "product_type": "SLC",
        "mission_data_take_id": 205463,
        "relative_orbit_number": 168,
        "start_time": "2021-04-01T05:26:22.396989",
        "stop_time": "2021-04-01T05:26:50.325833",
        "swaths": ["IW1", "IW2", "IW3"],
        "transmitter_receiver_polarisations": ["VV", "VH"],
    },
    "S1A_S6_SLC__1SDV_20210402T115512_20210402T115535_037271_046407_39FD": {
        "ascending_node_time": "2021-04-02T11:17:22.132050",
        "family_name": "SENTINEL-1",
        "mode": "SM",
        "number": "A",
        "orbit_number": 37271,
        "pass": "DESCENDING",
        "product_type": "SLC",
        "mission_data_take_id": 287751,
        "relative_orbit_number": 99,
        "start_time": "2021-04-02T11:55:12.030410",
        "stop_time": "2021-04-02T11:55:35.706705",
        "swaths": ["S6"],
        "transmitter_receiver_polarisations": ["VV", "VH"],
    },
    "S1A_S3_SLC__1SDV_20210401T152855_20210401T152914_037258_04638E_6001": {
        "ascending_node_time": "2021-04-01T13:53:42.874198",
        "family_name": "SENTINEL-1",
        "mode": "SM",
        "number": "A",
        "orbit_number": 37258,
        "pass": "ASCENDING",
        "product_type": "SLC",
        "mission_data_take_id": 287630,
        "relative_orbit_number": 86,
        "start_time": "2021-04-01T15:28:55.111501",
        "stop_time": "2021-04-01T15:29:14.277650",
        "swaths": ["S3"],
        "transmitter_receiver_polarisations": ["VV", "VH"],
    },
    "S1A_EW_SLC__1SDH_20210403T122536_20210403T122630_037286_046484_8152": {
        "ascending_node_time": "2021-04-03T11:58:30.792178",
        "family_name": "SENTINEL-1",
        "mode": "EW",
        "number": "A",
        "orbit_number": 37286,
        "pass": "DESCENDING",
        "product_type": "SLC",
        "mission_data_take_id": 287876,
        "relative_orbit_number": 114,
        "start_time": "2021-04-03T12:25:36.505937",
        "stop_time": "2021-04-03T12:26:30.902216",
        "swaths": ["EW1", "EW2", "EW3", "EW4", "EW5"],
        "transmitter_receiver_polarisations": ["HH", "HV"],
    },
    "S1B_WV_SLC__1SSV_20210403T083025_20210403T084452_026300_032390_D542": {
        "ascending_node_time": "2021-04-03T07:50:57.437371",
        "family_name": "SENTINEL-1",
        "mode": "WV",
        "number": "B",
        "orbit_number": 26300,
        "pass": "DESCENDING",
        "product_type": "SLC",
        "mission_data_take_id": 205712,
        "relative_orbit_number": 24,
        "start_time": "2021-04-03T08:30:25.749829",
        "stop_time": "2021-04-03T08:44:52.841818",
        "swaths": ["WV1", "WV2"],
        "transmitter_receiver_polarisations": ["VV"],
    },
    "S1B_IW_GRDH_1SDV_20210401T052623_20210401T052648_026269_032297_ECC8": {
        "ascending_node_time": "2021-04-01T04:49:55.637823",
        "family_name": "SENTINEL-1",
        "mode": "IW",
        "number": "B",
        "orbit_number": 26269,
        "pass": "DESCENDING",
        "product_type": "GRD",
        "mission_data_take_id": 205463,
        "relative_orbit_number": 168,
        "start_time": "2021-04-01T05:26:23.794457",
        "stop_time": "2021-04-01T05:26:48.793373",
        "swaths": ["IW"],
        "transmitter_receiver_polarisations": ["VV", "VH"],
    },
}

ANNOTATION_PATH = str(
    DATA_FOLDER
    / "S1B_IW_SLC__1SDV_20210401T052622_20210401T052650_026269_032297_EFA4.SAFE"
    / "annotation"
    / "s1b-iw1-slc-vv-20210401t052624-20210401t052649-026269-032297-004.xml"
)


def test_cached_sentinel1_schemas() -> None:
    res = esa_safe.cached_sentinel1_schemas("annotation")

    assert isinstance(res, xmlschema.XMLSchema)


def test_parse_tag() -> None:
    expected = {
        "timelinessCategory",
        "platformHeading",
        "radarFrequency",
        "rangeSamplingRate",
        "projection",
        "pass",
        "azimuthSteeringRate",
    }

    res = esa_safe.parse_tag(ANNOTATION_PATH, "//productInformation")

    assert isinstance(res, dict)
    assert set(res) == expected


def test_parse_tag_as_list() -> None:
    expected = {
        "azimuthTime",
        "firstValidSample",
        "sensingTime",
        "lastValidSample",
        "byteOffset",
        "azimuthAnxTime",
    }

    res = esa_safe.parse_tag_as_list(ANNOTATION_PATH, "//burst")

    assert isinstance(res, list)
    assert set(res[0]) == expected

    # XPath to a single element
    res = esa_safe.parse_tag_as_list(ANNOTATION_PATH, "//burst[1]")

    assert isinstance(res, list)
    assert set(res[0]) == expected

    # XPath to a non existent element
    res = esa_safe.parse_tag_as_list(ANNOTATION_PATH, "//dummy")

    assert isinstance(res, list)
    assert res == []


def test_parse_annotation_filename() -> None:
    res = esa_safe.parse_annotation_filename(
        "s1b-iw1-slc-vv-20210401t052624-20210401t052649-026269-032297-004.xml"
    )

    assert res == ("", "iw1", "vv", "20210401t052624")

    with pytest.raises(ValueError):
        esa_safe.parse_annotation_filename("")


def test_findtext() -> None:
    tree = ElementTree.fromstring("<root><child>text</child></root>")

    res = esa_safe.findtext(tree, ".//child")

    assert res == "text"

    with pytest.raises(ValueError):
        esa_safe.findtext(tree, ".//dummy")


def test_findall() -> None:
    tree = ElementTree.fromstring("<root><c1>text</c1><c2></c2></root>")

    res = esa_safe.findall(tree, ".//c1")

    assert res == ["text"]

    with pytest.raises(ValueError):
        esa_safe.findall(tree, ".//c2")


@pytest.mark.parametrize("product_id,expected", SENTINEL1_ATTRIBUTES.items())
def test_parse_manifest_sentinel1(product_id: str, expected: Dict[str, Any]) -> None:
    manifest_path = DATA_FOLDER / (product_id + ".SAFE") / "manifest.safe"

    res_attrs, res_files = esa_safe.parse_manifest_sentinel1(manifest_path)

    assert res_attrs == expected


def test_make_stac_item() -> None:
    attrs = SENTINEL1_ATTRIBUTES[
        "S1B_IW_GRDH_1SDV_20210401T052623_20210401T052648_026269_032297_ECC8"
    ]
    expected = {
        "constellation": "sentinel-1",
        "instrument": ["c-sar"],
        "platform": "sentinel-1b",
        "sat:orbit_state": "descending",
        "sat:absolute_orbit": 26269,
        "sat:relative_orbit": 168,
        "sat:anx_datetime": "2021-04-01T04:49:55.637823Z",
        "sar:instrument_mode": "IW",
        "sar:frequency_band": "C",
        "sar:polarizations": ["VV", "VH"],
        "sar:product_type": "GRD",
        "sar:observation_direction": "right",
    }

    assert esa_safe.make_stac_item(attrs) == expected



import functools
import importlib.resources
import os
import re
from typing import Any, Mapping, TextIO, Union
from xml.etree import ElementTree

import xmlschema

PathType = Union[str, "os.PathLike[str]"]
PathOrFileType = Union[PathType, TextIO]


SENTINEL1_NAMESPACES = {
    "safe": "http://www.esa.int/safe/sentinel-1.0",
    "s1": "http://www.esa.int/safe/sentinel-1.0/sentinel-1",
    "s1sarl1": "http://www.esa.int/safe/sentinel-1.0/sentinel-1/sar/level-1",
    "gml": "http://www.opengis.net/gml",
}

SENTINEL1_FOLDER = importlib.resources.files("xarray_sentinel") / "resources/sentinel1"
SENTINEL1_SCHEMAS = {
    "manifest": SENTINEL1_FOLDER / "my-xfdu.xsd",
    "annotation": SENTINEL1_FOLDER / "s1-level-1-product.xsd",
    "calibration": SENTINEL1_FOLDER / "s1-level-1-calibration.xsd",
    "noise": SENTINEL1_FOLDER / "s1-level-1-noise.xsd",
    "aux_orbit": SENTINEL1_FOLDER / "my-schema_orb.xsd",
}

SENTINEL2_NAMESPACES = {
    "safe": "http://www.esa.int/safe/sentinel/1.1",
}


@functools.lru_cache
def cached_sentinel1_schemas(schema_type: str) -> xmlschema.XMLSchema:
    return xmlschema.XMLSchema(str(SENTINEL1_SCHEMAS[schema_type]))


def parse_tag(
    xml_path: PathOrFileType,
    query: str,
    schema_type: str = "annotation",
    validation: str = "skip",
) -> dict[str, Any]:
    schema = cached_sentinel1_schemas(schema_type)
    if hasattr(xml_path, "seek"):
        xml_path.seek(0)
    xml_tree = ElementTree.parse(xml_path)
    tag_dict: Any = schema.decode(xml_tree, query, validation=validation)
    assert isinstance(tag_dict, dict), f"{type(tag_dict)} is not dict"
    return tag_dict


def parse_tag_as_list(
    xml_path: PathOrFileType,
    query: str,
    schema_type: str = "annotation",
    validation: str = "skip",
) -> list[dict[str, Any]]:
    schema = cached_sentinel1_schemas(schema_type)
    xml_tree = ElementTree.parse(xml_path)
    tag: Any = schema.decode(xml_tree, query, validation=validation)
    if tag is None:
        tag = []
    elif isinstance(tag, dict):
        tag = [tag]
    tag_list: list[dict[str, Any]] = tag
    assert isinstance(tag_list, list), f"{type(tag_list)} is not list"
    return tag_list


def findtext(
    tree: ElementTree.Element,
    query: str,
    namespaces: dict[str, str] = SENTINEL1_NAMESPACES,
) -> str:
    value = tree.findtext(query, namespaces=namespaces)
    if value is None:
        raise ValueError(f"{query=} returned None")
    return value


def findall(
    tree: ElementTree.Element,
    query: str,
    namespaces: dict[str, str] = SENTINEL1_NAMESPACES,
) -> list[str]:
    tags = tree.findall(query, namespaces=namespaces)
    values: list[str] = []
    for tag in tags:
        if tag.text is None:
            raise ValueError(f"{query=} returned None")
        values.append(tag.text)
    return values


def parse_annotation_filename(name: str) -> tuple[str, str, str, str]:
    match = re.match(
        r"([a-z-]*)s1[abcd]-([^-]*)-[^-]*-([^-]*)-([\dt]*)-", os.path.basename(name)
    )
    if match is None:
        raise ValueError(f"cannot parse name {name!r}")
    return tuple(match.groups())  # type: ignore


@functools.lru_cache
def parse_manifest_sentinel1(
    manifest_path: PathOrFileType,
) -> tuple[dict[str, Any], dict[str, tuple[str, str, str, str, str]]]:
    # We use ElementTree because we didn't find a XSD definition for the manifest
    manifest = ElementTree.parse(manifest_path).getroot()

    family_name = findtext(manifest, ".//safe:platform/safe:familyName")
    if family_name != "SENTINEL-1":
        raise ValueError(f"{family_name=} not supported")

    number = findtext(manifest, ".//safe:platform/safe:number")
    mode = findtext(manifest, ".//s1sarl1:instrumentMode/s1sarl1:mode")
    swaths = findall(manifest, ".//s1sarl1:instrumentMode/s1sarl1:swath")

    orbit_number = findall(manifest, ".//safe:orbitNumber")
    if len(orbit_number) != 2 or orbit_number[0] != orbit_number[1]:
        raise ValueError(f"{orbit_number=} not supported")

    relative_orbit_number = findall(manifest, ".//safe:relativeOrbitNumber")
    if (
        len(relative_orbit_number) != 2
        or relative_orbit_number[0] != relative_orbit_number[1]
    ):
        raise ValueError(f"{relative_orbit_number=} not supported")

    orbit_pass = findtext(manifest, ".//s1:pass")
    if orbit_pass not in {"ASCENDING", "DESCENDING"}:
        raise ValueError(f"pass={orbit_pass} not supported")

    ascending_node_time = findtext(manifest, ".//s1:ascendingNodeTime")

    mission_data_take_id = findtext(manifest, ".//s1sarl1:missionDataTakeID")
    transmitter_receiver_polarisations = findall(
        manifest, ".//s1sarl1:transmitterReceiverPolarisation"
    )
    product_type = findtext(manifest, ".//s1sarl1:productType")

    start_time = findtext(manifest, ".//safe:startTime")
    stop_time = findtext(manifest, ".//safe:stopTime")

    attributes = {
        "family_name": family_name,
        "number": number,
        "mode": mode,
        "swaths": swaths,
        "orbit_number": int(orbit_number[0]),
        "relative_orbit_number": int(relative_orbit_number[0]),
        "pass": orbit_pass,
        "ascending_node_time": ascending_node_time,
        "mission_data_take_id": int(mission_data_take_id),
        "transmitter_receiver_polarisations": transmitter_receiver_polarisations,
        "product_type": product_type,
        "start_time": start_time,
        "stop_time": stop_time,
    }

    files = {}

    for file_tag in manifest.findall(".//dataObjectSection/dataObject"):
        location_tag = file_tag.find(".//fileLocation")
        if location_tag is not None:
            file_href = location_tag.attrib["href"]
            try:
                description = parse_annotation_filename(os.path.basename(file_href))
            except ValueError:
                continue
            file_type = file_tag.attrib["repID"]
            files[file_href] = (file_type,) + description

    return attributes, files


# DEPRECATED
def make_stac_item(attrs: Mapping[str, Any]) -> dict[str, Any]:
    assert attrs["family_name"] == "SENTINEL-1"

    stac_item = {
        "constellation": "sentinel-1",
        "platform": "sentinel-1" + attrs["number"].lower(),
        "instrument": ["c-sar"],
        "sat:orbit_state": attrs["pass"].lower(),
        "sat:absolute_orbit": attrs["orbit_number"],
        "sat:relative_orbit": attrs["relative_orbit_number"],
        "sat:anx_datetime": attrs["ascending_node_time"] + "Z",
        "sar:instrument_mode": attrs["mode"],
        "sar:frequency_band": "C",
        "sar:polarizations": attrs["transmitter_receiver_polarisations"],
        "sar:product_type": attrs["product_type"],
        "sar:observation_direction": "right",
    }
    return stac_item



import pathlib

import xarray as xr

from xarray_sentinel import eopf_product

DATA_FOLDER = pathlib.Path(__file__).parent / "data"

SLC_S3 = (
    DATA_FOLDER
    / "S1A_S3_SLC__1SDV_20210401T152855_20210401T152914_037258_04638E_6001.SAFE"
)


def test_open_datatree() -> None:
    res = eopf_product.open_datatree(SLC_S3, check_files_exist=True)

    assert isinstance(res, xr.DataTree)



"""Map Sentinel-1 data products to xarray.

References
----------
  - Sentinel-1 document library:
    https://sentinels.copernicus.eu/web/sentinel/user-guides/sentinel-1-sar/document-library
  - Sentinel-1 Product Specification v3.9 07 May 2021 S1-RS-MDA-52-7441-3-9 documenting IPF 3.40
    https://sentinel.esa.int/documents/247904/1877131/S1-RS-MDA-52-7441-3-9-2_Sentinel-1ProductSpecification.pdf
  - Sentinel-1 Product Specification v3.7 27 February 2020 S1-RS-MDA-52-7441 documenting IPF 3.30
    https://sentinel.esa.int/documents/247904/1877131/Sentinel-1-Product-Specification
"""

from __future__ import annotations

import os
import warnings
from typing import Any, Sequence, TypeVar
from xml.etree import ElementTree

import fsspec
import numpy as np
import numpy.typing as npt
import pandas as pd
import rasterio
import xarray as xr

from . import conventions, eopf_metadata, esa_safe

SPEED_OF_LIGHT = 299_792_458  # m / s
ONE_SECOND = np.timedelta64(1, "s")


DataArrayOrDataset = TypeVar("DataArrayOrDataset", xr.DataArray, xr.Dataset)


def get_fs_path(
    urlpath_or_path: esa_safe.PathType,
    fs: fsspec.AbstractFileSystem | None = None,
    storage_options: dict[str, Any] | None = None,
) -> tuple[fsspec.AbstractFileSystem, str]:
    if fs is not None and storage_options is not None:
        raise TypeError("only one of 'fs' and 'storage_options' can be not None")

    if fs is None:
        fs, _, paths = fsspec.get_fs_token_paths(
            urlpath_or_path, storage_options=storage_options
        )
        if len(paths) == 0:
            raise ValueError(f"file or object not found {urlpath_or_path!r}")
        elif len(paths) > 1:
            raise ValueError(f"multiple files or objects found {urlpath_or_path!r}")
        path = paths[0]
    else:
        path = str(urlpath_or_path)

    if fs.isdir(path):
        path = os.path.join(path, "manifest.safe")

    return fs, path


def normalise_group(group: str | None) -> tuple[str, int | None]:
    if group is None:
        group = ""
    if group.startswith("/"):
        group = group[1:]
    burst_index = None
    parent_group, _, last_name = group.rpartition("/")
    if parent_group.count("/") == 1 and last_name.isnumeric():
        burst_index = int(last_name)
        group = parent_group
    return group, burst_index


def open_calibration_dataset(
    calibration: esa_safe.PathType, attrs: dict[str, Any] = {}
) -> xr.Dataset:
    calibration_vectors = esa_safe.parse_tag_as_list(
        calibration, "//calibrationVector", "calibration"
    )
    cal_attrs = esa_safe.parse_tag(
        calibration, "//calibrationInformation", "calibration"
    )
    attrs["absoluteCalibrationConstant"] = cal_attrs["absoluteCalibrationConstant"]
    azimuth_time_list = []
    pixel_list = []
    line_list = []
    sigmaNought_list = []
    betaNought_list = []
    gamma_list = []
    dn_list = []
    for vector in calibration_vectors:
        azimuth_time_list.append(vector["azimuthTime"])
        line_list.append(vector["line"])
        pixel = np.fromstring(vector["pixel"]["$"], dtype=int, sep=" ")
        pixel_list.append(pixel)
        sigmaNought = np.fromstring(
            vector["sigmaNought"]["$"], dtype=np.float32, sep=" "
        )
        sigmaNought_list.append(sigmaNought)
        betaNought = np.fromstring(vector["betaNought"]["$"], dtype=np.float32, sep=" ")
        betaNought_list.append(betaNought)
        gamma = np.fromstring(vector["gamma"]["$"], dtype=np.float32, sep=" ")
        gamma_list.append(gamma)
        dn = np.fromstring(vector["dn"]["$"], dtype=np.float32, sep=" ")
        dn_list.append(dn)

    pixel = np.array(pixel_list)
    if (pixel - pixel[0]).any():
        raise ValueError(
            "Unable to organise calibration vectors in a regular line-pixel grid"
        )
    data_vars = {
        "azimuth_time": ("line", [np.datetime64(dt, "ns") for dt in azimuth_time_list]),
        "sigmaNought": (("line", "pixel"), sigmaNought_list),
        "betaNought": (("line", "pixel"), betaNought_list),
        "gamma": (("line", "pixel"), gamma_list),
        "dn": (("line", "pixel"), dn_list),
    }
    coords = {"line": line_list, "pixel": pixel_list[0]}

    return xr.Dataset(data_vars=data_vars, coords=coords, attrs=attrs)


def open_reference_replica_dataset(
    annotation_path: esa_safe.PathType, attrs: dict[str, Any] = {}
) -> xr.Dataset:
    reference_replica = esa_safe.parse_tag_as_list(
        annotation_path, "//replicaInformationList/replicaInformation/referenceReplica"
    )[0]
    attrs.update(
        {
            "azimuth_time": reference_replica["azimuthTime"],
            "chirpSource": reference_replica["chirpSource"],
            "pgSource": reference_replica["pgSource"],
            "timeDelay": reference_replica["timeDelay"],
            "gain_re": reference_replica["gain"]["re"],
            "gain_im": reference_replica["gain"]["im"],
        }
    )

    reference_replica_amplitude_coefficients = [
        float(v) for v in reference_replica["amplitudeCoefficients"]["$"].split()
    ]

    reference_replica_phase_coefficients = [
        float(v) for v in reference_replica["phaseCoefficients"]["$"].split()
    ]

    coords: dict[str, Any] = {
        "degree": range(len(reference_replica_amplitude_coefficients))
    }
    data_vars = {
        "amplitudeCoefficients": (("degree"), reference_replica_amplitude_coefficients),
        "phaseCoefficients": (("degree"), reference_replica_phase_coefficients),
    }
    da = xr.Dataset(data_vars=data_vars, coords=coords, attrs=attrs)
    return da


def open_antenna_pattern(
    annotation_path: esa_safe.PathType, attrs: dict[str, Any] = {}
) -> xr.Dataset:
    antenna_pattern_list = esa_safe.parse_tag_as_list(
        annotation_path, "//antennaPattern/antennaPatternList/antennaPattern"
    )

    slant_range_time_list = []
    azimuth_time_list = []
    elevation_angle_list = []
    elevation_pattern_list = []
    incidence_angle_list = []
    terrain_height_list = []
    roll_list = []

    for vector in antenna_pattern_list:
        azimuth_time_list.append(vector["azimuthTime"])
        slant_range_time = np.fromstring(
            vector["slantRangeTime"]["$"], dtype=np.float32, sep=" "
        )
        slant_range_time_list.append(slant_range_time)

        elevation_angle = np.fromstring(
            vector["elevationAngle"]["$"], dtype=np.float32, sep=" "
        )
        elevation_angle_list.append(elevation_angle)

        elevation_pattern = np.fromstring(
            vector["elevationPattern"]["$"], dtype=np.float32, sep=" "
        )
        elevation_pattern_list.append(
            elevation_pattern[::2] * 1j + elevation_pattern[1::2]
        )

        incidence_angle = np.fromstring(
            vector["incidenceAngle"]["$"], dtype=np.float32, sep=" "
        )
        incidence_angle_list.append(incidence_angle)

        terrain_height_list.append(vector["terrainHeight"])
        roll_list.append(vector["roll"])

    slant_range_time_array = np.array(slant_range_time_list)
    if (slant_range_time_array - slant_range_time_array[0]).any():
        raise ValueError(
            "Unable to organise noise vectors in a regular line-pixel grid"
        )
    data_vars = {
        "elevationAngle": (
            ("azimuth_time", "slant_range_time"),
            np.array(elevation_angle_list),
        ),
        # "elevationPattern": (
        #     ( "azimuth_time", "slant_range_time"), np.array(elevation_pattern_list)),
        "incidenceAngle": (
            ("azimuth_time", "slant_range_time"),
            np.array(incidence_angle_list),
        ),
        "terrainHeight": ("azimuth_time", terrain_height_list),
        "roll": ("azimuth_time", roll_list),
    }
    coords = {
        "slant_range_time": slant_range_time_array[0],
        "azimuth_time": [np.datetime64(dt, "ns") for dt in azimuth_time_list],
    }
    da = xr.Dataset(data_vars=data_vars, coords=coords, attrs=attrs)
    return da


def open_replica_dataset(
    annotation_path: esa_safe.PathType, attrs: dict[str, Any] = {}
) -> xr.Dataset:
    replicaList = esa_safe.parse_tag_as_list(
        annotation_path,
        "//replicaInformationList/replicaInformation/replicaList/replica",
    )
    azimuth_time_list = []
    cross_correlation_bandwidth_list = []
    cross_correlation_pslr_list = []
    cross_correlation_peak_location_list = []
    reconstructed_replica_valid_flag_list = []
    pg_product_phase_list = []
    pg_product_amplitude_list = []
    model_pg_product_amplitude_list = []
    model_pg_product_phase_list = []
    relative_pg_product_valid_flag_list = []
    absolute_pg_product_valid_flag_list = []
    internal_time_delay_list = []

    for replica in replicaList:
        azimuth_time_list.append(replica["azimuthTime"])
        cross_correlation_bandwidth_list.append(replica["crossCorrelationBandwidth"])
        cross_correlation_pslr_list.append(replica["crossCorrelationPslr"])
        cross_correlation_peak_location_list.append(
            replica["crossCorrelationPeakLocation"]
        )
        reconstructed_replica_valid_flag_list.append(
            replica["reconstructedReplicaValidFlag"]
        )
        pg_product_amplitude_list.append(replica["pgProductAmplitude"])
        pg_product_phase_list.append(replica["pgProductPhase"])
        model_pg_product_amplitude_list.append(replica["modelPgProductAmplitude"])
        model_pg_product_phase_list.append(replica["modelPgProductPhase"])
        relative_pg_product_valid_flag_list.append(
            replica["relativePgProductValidFlag"]
        )
        absolute_pg_product_valid_flag_list.append(
            replica["absolutePgProductValidFlag"]
        )
        internal_time_delay_list.append(replica["internalTimeDelay"])

    coords: dict[str, Any] = {
        "azimuth_time": [np.datetime64(dt, "ns") for dt in azimuth_time_list],
    }
    data_vars = {
        "crossCorrelationBandwidth": (
            ("azimuth_time"),
            cross_correlation_bandwidth_list,
        ),
        "crossCorrelationPslr": (("azimuth_time"), cross_correlation_pslr_list),
        "crossCorrelationPeakLocation": (
            ("azimuth_time"),
            cross_correlation_peak_location_list,
        ),
        "reconstructedReplicaValidFlag": (
            ("azimuth_time"),
            reconstructed_replica_valid_flag_list,
        ),
        "pgProductAmplitude": (("azimuth_time"), pg_product_amplitude_list),
        "pgProductPhase": (("azimuth_time"), pg_product_phase_list),
        "modelPgProductAmplitude": (("azimuth_time"), model_pg_product_amplitude_list),
        "modelPgProductPhase": (("azimuth_time"), model_pg_product_phase_list),
        "relativePgProductValidFlag": (
            ("azimuth_time"),
            relative_pg_product_valid_flag_list,
        ),
        "absolutePgProductValidFlag": (
            ("azimuth_time"),
            absolute_pg_product_valid_flag_list,
        ),
        "internalTimeDelay": (("azimuth_time"), internal_time_delay_list),
    }

    da = xr.Dataset(data_vars=data_vars, coords=coords, attrs=attrs)
    return da


def open_noise_range_dataset(
    noise: esa_safe.PathType, attrs: dict[str, Any] = {}
) -> xr.Dataset:
    noise_vectors = esa_safe.parse_tag_as_list(noise, "//noiseRangeVector", "noise")

    azimuth_time_list = []
    pixel_list = []
    line_list = []
    noiseRangeLut_list = []
    for vector in noise_vectors:
        azimuth_time_list.append(vector["azimuthTime"])
        line_list.append(vector["line"])
        pixel = np.fromstring(vector["pixel"]["$"], dtype=int, sep=" ")
        pixel_list.append(pixel)
        noiseRangeLut = np.fromstring(
            vector["noiseRangeLut"]["$"], dtype=np.float32, sep=" "
        )
        noiseRangeLut_list.append(noiseRangeLut)

    pixel = np.array(pixel_list)
    if (pixel - pixel[0]).any():
        raise ValueError(
            "Unable to organise noise vectors in a regular line-pixel grid"
        )
    data_vars = {
        "azimuth_time": ("line", [np.datetime64(dt, "ns") for dt in azimuth_time_list]),
        "noiseRangeLut": (("line", "pixel"), noiseRangeLut_list),
    }
    coords = {"line": line_list, "pixel": pixel_list[0]}

    return xr.Dataset(data_vars=data_vars, coords=coords, attrs=attrs)


def open_noise_azimuth_dataset(
    noise: esa_safe.PathType, attrs: dict[str, Any] = {}
) -> xr.Dataset:
    noise_vectors = esa_safe.parse_tag_as_list(noise, "//noiseAzimuthVector", "noise")

    first_range_sample = []
    line_list = []
    noise_azimuth_lut_list = []
    for vector in noise_vectors:
        first_range_sample.append(vector["firstRangeSample"])
        line = np.fromstring(vector["line"]["$"], dtype=int, sep=" ")
        line_list.append(line)
        noise_azimuth_lut = np.fromstring(
            vector["noiseAzimuthLut"]["$"], dtype=np.float32, sep=" "
        )
        noise_azimuth_lut_list.append(noise_azimuth_lut)

    # BROKEN: GRDs have line and noise_azimuth_lut of different size, we take the first one
    data_vars = {}
    coords = {}
    if first_range_sample:
        data_vars["noiseAzimuthLut"] = ("line", noise_azimuth_lut_list[0])
        coords["line"] = line_list[0]

    return xr.Dataset(data_vars=data_vars, coords=coords, attrs=attrs)


def open_coordinate_conversion_dataset(
    annotation_path: esa_safe.PathType, attrs: dict[str, Any] = {}
) -> xr.Dataset:
    coordinate_conversion = esa_safe.parse_tag_as_list(
        annotation_path, "//coordinateConversionList/coordinateConversion"
    )
    if len(coordinate_conversion) == 0:
        raise TypeError("coordinateConversion tag not present in annotations")

    gr0 = []
    sr0 = []
    azimuth_time = []
    slant_range_time = []
    srgrCoefficients: list[list[float]] = []
    grsrCoefficients: list[list[float]] = []
    for values in coordinate_conversion:
        sr0.append(values["sr0"])
        gr0.append(values["gr0"])
        azimuth_time.append(values["azimuthTime"])
        slant_range_time.append(values["slantRangeTime"])
        srgrCoefficients.append(
            [float(v) for v in values["srgrCoefficients"]["$"].split()]
        )
        grsrCoefficients.append(
            [float(v) for v in values["grsrCoefficients"]["$"].split()]
        )

    coords: dict[str, Any] = {}
    data_vars: dict[str, Any] = {}
    coords["azimuth_time"] = [np.datetime64(dt, "ns") for dt in azimuth_time]
    coords["degree"] = list(range(len(srgrCoefficients[0])))

    data_vars["gr0"] = ("azimuth_time", gr0)
    data_vars["sr0"] = ("azimuth_time", sr0)
    data_vars["slant_range_time"] = ("azimuth_time", slant_range_time)
    data_vars["srgrCoefficients"] = (("azimuth_time", "degree"), srgrCoefficients)
    data_vars["grsrCoefficients"] = (("azimuth_time", "degree"), grsrCoefficients)

    return xr.Dataset(data_vars=data_vars, coords=coords, attrs=attrs)


# taken from `np.cross` docs
def cross2d(x: np.ndarray, y: np.ndarray) -> float:
    return float(x[..., 0] * y[..., 1] - x[..., 1] * y[..., 0])


def is_clockwise(poly: list[tuple[float, float]]) -> bool:
    start = np.array(poly[0])
    return cross2d(poly[1] - start, poly[2] - start) < 0


def open_gcp_dataset(
    annotation: esa_safe.PathOrFileType, attrs: dict[str, Any] = {}
) -> xr.Dataset:
    geolocation_grid_points = esa_safe.parse_tag_as_list(
        annotation, "//geolocationGridPoint"
    )

    azimuth_time = []
    slant_range_time = []
    line_set = set()
    pixel_set = set()
    for ggp in geolocation_grid_points:
        if ggp["line"] not in line_set:
            azimuth_time.append(np.datetime64(ggp["azimuthTime"], "ns"))
            line_set.add(ggp["line"])
        if ggp["pixel"] not in pixel_set:
            slant_range_time.append(ggp["slantRangeTime"])
            pixel_set.add(ggp["pixel"])
    shape = (len(azimuth_time), len(slant_range_time))
    dims = ("azimuth_time", "slant_range_time")
    data_vars = {
        "latitude": (dims, np.full(shape, np.nan), attrs),
        "longitude": (dims, np.full(shape, np.nan), attrs),
        "height": (dims, np.full(shape, np.nan), attrs),
        "incidenceAngle": (dims, np.full(shape, np.nan), attrs),
        "elevationAngle": (dims, np.full(shape, np.nan), attrs),
    }
    line = sorted(line_set)
    pixel = sorted(pixel_set)
    for ggp in geolocation_grid_points:
        for var in data_vars:
            j = line.index(ggp["line"])
            i = pixel.index(ggp["pixel"])
            data_vars[var][1][j, i] = ggp[var]

    ds = xr.Dataset(
        data_vars=data_vars,
        coords={
            "azimuth_time": azimuth_time,
            "slant_range_time": slant_range_time,
            "line": ("azimuth_time", line),
            "pixel": ("slant_range_time", pixel),
        },
        attrs=attrs,
    )

    footprint = get_footprint_linestring(ds.azimuth_time, ds.slant_range_time, ds)
    geospatial_attrs = make_geospatial_attributes(footprint)
    ds.attrs.update(geospatial_attrs)

    return ds


def get_footprint_linestring(
    azimuth_time: xr.DataArray,
    slant_range_time: xr.DataArray,
    gcp: xr.Dataset,
    method: xr.core.types.InterpOptions = "linear",
    kwargs: dict[str, Any] = {"fill_value": "extrapolate"},
) -> list[tuple[float, float]]:
    azimuth_time_mm = [azimuth_time.min(), azimuth_time.max()]
    slant_range_time_mm = [slant_range_time.min(), slant_range_time.max()]

    footprint = []
    for j, i in [(0, 0), (1, 0), (1, 1), (0, 1)]:
        lat = float(
            gcp["latitude"].interp(
                azimuth_time=azimuth_time_mm[j],
                slant_range_time=slant_range_time_mm[i],
                method=method,
                kwargs=kwargs,
            )
        )
        lon = float(
            gcp["longitude"].interp(
                azimuth_time=azimuth_time_mm[j],
                slant_range_time=slant_range_time_mm[i],
                method=method,
                kwargs=kwargs,
            )
        )
        footprint.append((lon, lat))

    # check that the poly as the correct orientation
    if is_clockwise(footprint):
        footprint = footprint[::-1]

    # close the polygon
    footprint.append(footprint[0])

    return footprint


def make_geospatial_attributes(
    footprint: Sequence[tuple[float, float]],
) -> dict[str, Any]:
    wkt = "POLYGON((" + ",".join(f"{y} {x}" for y, x in footprint) + "))"
    geospatial_attrs = {
        "geospatial_bounds": wkt,
        "geospatial_lat_min": min(lat for _, lat in footprint),
        "geospatial_lat_max": max(lat for _, lat in footprint),
        "geospatial_lon_min": min(lon for lon, _ in footprint),
        "geospatial_lon_max": max(lon for lon, _ in footprint),
    }
    return geospatial_attrs


def open_attitude_dataset(
    annotation: esa_safe.PathOrFileType, attrs: dict[str, Any] = {}
) -> xr.Dataset:
    attitudes = esa_safe.parse_tag_as_list(annotation, "//attitude")

    variables = ["q0", "q1", "q2", "q3", "wx", "wy", "wz", "pitch", "roll", "yaw"]
    azimuth_time: list[Any] = []
    data_vars: dict[str, Any]
    data_vars = {var: ("azimuth_time", [], attrs) for var in variables}
    for attitude in attitudes:
        azimuth_time.append(attitude["time"])
        for var in variables:
            data_vars[var][1].append(attitude[var])

    ds = xr.Dataset(
        data_vars=data_vars,
        coords={"azimuth_time": [np.datetime64(dt, "ns") for dt in azimuth_time]},
        attrs=attrs,
    )

    return ds


def make_orbit(
    azimuth_time: list[Any],
    positions: list[list[Any]],
    velocities: list[list[Any]],
    attrs: dict[str, Any] = {},
) -> xr.Dataset:
    position = xr.Variable(data=positions, dims=("axis", "azimuth_time"))
    velocity = xr.Variable(data=velocities, dims=("axis", "azimuth_time"))

    ds = xr.Dataset(
        data_vars={"position": position, "velocity": velocity},
        attrs=attrs,
        coords={
            "azimuth_time": [np.datetime64(dt, "ns") for dt in azimuth_time],
            "axis": [0, 1, 2],
        },
    )
    for data_var in ds.data_vars:
        ds[data_var].attrs = attrs

    return ds


def open_orbit_dataset(
    annotation: esa_safe.PathOrFileType, attrs: dict[str, Any] = {}
) -> xr.Dataset:
    orbits = esa_safe.parse_tag_as_list(annotation, "//orbit")

    attrs = attrs.copy()
    reference_system = orbits[0]["frame"]
    if reference_system is not None:
        attrs.update({"reference_system": reference_system})

    azimuth_times: list[Any] = []
    positions: list[list[Any]] = [[], [], []]
    velocities: list[list[Any]] = [[], [], []]
    for orbit in orbits:
        azimuth_times.append(orbit["time"])
        positions[0].append(orbit["position"]["x"])
        positions[1].append(orbit["position"]["y"])
        positions[2].append(orbit["position"]["z"])
        velocities[0].append(orbit["velocity"]["x"])
        velocities[1].append(orbit["velocity"]["y"])
        velocities[2].append(orbit["velocity"]["z"])
        if orbit["frame"] != reference_system:
            warnings.warn(
                "reference_system is not consistent in all the state vectors. "
            )
            reference_system = None

    return make_orbit(azimuth_times, positions, velocities, attrs)


def open_dc_estimate_dataset(
    annotation: esa_safe.PathOrFileType, attrs: dict[str, Any] = {}
) -> xr.Dataset:
    dc_estimates = esa_safe.parse_tag_as_list(annotation, "//dcEstimate")

    azimuth_time = []
    t0 = []
    data_dc_poly = []
    geometry_dc_poly = []
    data_dc_rms_error = []
    data_dc_rms_error_above_threshold = []
    fine_dce_azimuth_start_time = []
    fine_dce_azimuth_stop_time = []
    for dc_estimate in dc_estimates:
        geometry_dc_poly.append(
            [float(c) for c in dc_estimate["geometryDcPolynomial"]["$"].split()]
        )
        data_dc_rms_error.append(dc_estimate["dataDcRmsError"])
        data_dc_rms_error_above_threshold.append(
            dc_estimate["dataDcRmsErrorAboveThreshold"]
        )
        fine_dce_azimuth_start_time.append(dc_estimate["fineDceAzimuthStartTime"])
        fine_dce_azimuth_stop_time.append(dc_estimate["fineDceAzimuthStopTime"])
        azimuth_time.append(dc_estimate["azimuthTime"])
        t0.append(dc_estimate["t0"])
        data_dc_poly.append(
            [float(c) for c in dc_estimate["dataDcPolynomial"]["$"].split()]
        )

    ds = xr.Dataset(
        data_vars={
            "t0": ("azimuth_time", t0, attrs),
            "data_dc_polynomial": (("azimuth_time", "degree"), data_dc_poly, attrs),
            "geometry_dc_polynomial": (
                ("azimuth_time", "degree"),
                geometry_dc_poly,
                attrs,
            ),
            "data_dc_rms_error": ("azimuth_time", data_dc_rms_error, attrs),
            "data_dc_rms_error_above_threshold": (
                "azimuth_time",
                data_dc_rms_error,
                attrs,
            ),
            #
            # Not present in the new Zarr products
            #
            # "fine_dce_azimuth_start_time": (
            #     "azimuth_time",
            #     [np.datetime64(at, "ns") for at in fine_dce_azimuth_start_time],
            # ),
            # "fine_dce_azimuth_stop_time": (
            #     "azimuth_time",
            #     [np.datetime64(at, "ns") for at in fine_dce_azimuth_stop_time],
            # ),
        },
        coords={
            "azimuth_time": [np.datetime64(at, "ns") for at in azimuth_time],
            "degree": list(range(len(data_dc_poly[0]))),
        },
        attrs=attrs,
    )
    return ds


def open_azimuth_fm_rate_dataset(
    annotation: esa_safe.PathOrFileType, attrs: dict[str, Any] = {}
) -> xr.Dataset:
    azimuth_fm_rates = esa_safe.parse_tag_as_list(annotation, "//azimuthFmRate")

    azimuth_time = []
    t0 = []
    azimuth_fm_rate_poly = []
    for azimuth_fm_rate in azimuth_fm_rates:
        azimuth_time.append(azimuth_fm_rate["azimuthTime"])
        t0.append(azimuth_fm_rate["t0"])
        azimuth_fm_rate_poly.append(
            [float(c) for c in azimuth_fm_rate["azimuthFmRatePolynomial"]["$"].split()]
        )

    ds = xr.Dataset(
        data_vars={
            "t0": (
                "azimuth_time",
                t0,
                attrs,
            ),
            "azimuth_fm_rate_polynomial": (
                ("azimuth_time", "degree"),
                azimuth_fm_rate_poly,
                attrs,
            ),
        },
        coords={
            "azimuth_time": [np.datetime64(at, "ns") for at in azimuth_time],
            "degree": list(range(len(azimuth_fm_rate_poly[0]))),
        },
        attrs=attrs,
    )
    return ds


def find_available_groups(
    product_files: dict[str, tuple[str, str, str, str, str]],
    product_path: str,
    product_type: str,
    check_files_exist: bool = False,
    fs: fsspec.AbstractFileSystem = fsspec.filesystem("file"),
) -> dict[str, list[str]]:
    groups: dict[str, list[str]] = {}
    for path, (type, _, swath, polarization, _) in product_files.items():
        swath_pol_group = f"{swath}/{polarization}".upper()
        abspath = os.path.join(product_path, os.path.normpath(path))
        if check_files_exist:
            if not fs.exists(abspath):
                continue
        if type == "s1Level1ProductSchema":
            groups[swath.upper()] = [""]
            groups[swath_pol_group] = [abspath] + groups.get(swath_pol_group, [])
            for metadata_group in [
                "orbit",
                "attitude",
                "azimuth_fm_rate",
                "dc_estimate",
                "gcp",
                # "replica",
                # "reference_replica",
                # "antenna_pattern",
            ]:
                if product_type == "GRD" and metadata_group == "antenna_pattern":
                    continue
                groups[f"{swath_pol_group}/{metadata_group}"] = [abspath]
            if product_type == "GRD":
                groups[f"{swath_pol_group}/coordinate_conversion"] = [abspath]
        elif type == "s1Level1CalibrationSchema":
            groups[f"{swath_pol_group}/calibration"] = [abspath]
        elif type == "s1Level1NoiseSchema":
            groups[f"{swath_pol_group}/noise_range"] = [abspath]
            groups[f"{swath_pol_group}/noise_azimuth"] = [abspath]
        elif type == "s1Level1MeasurementSchema":
            groups[swath_pol_group] = [abspath] + groups.get(swath_pol_group, [])

    return groups


def open_rasterio_dataarray(
    measurement: esa_safe.PathOrFileType,
    fs: fsspec.AbstractFileSystem | None,
    chunks: dict[str, int] | None,
) -> xr.DataArray:
    # fsspec needs rasterio >= 1.3.0, but we allow earlier rasterio versions for local files
    if fs is None or isinstance(fs, fsspec.implementations.local.LocalFileSystem):
        try:
            arr = xr.open_dataarray(measurement, engine="rasterio", chunks=chunks)
        except rasterio.RasterioIOError as ex:
            if "No such file" in str(ex):
                raise FileNotFoundError(str(ex))
            raise
    else:
        # FIXME: rioxarray / rasterio do not support opening a file object any more, so
        #   full fsspec support via `fs.open(measurement)` is now broken and we fall back
        #   to hope that rasterio remote protocols works with fsspec urlpaths
        maybe_rasterio_urlpath = fs.unstrip_protocol(measurement)
        arr = xr.open_dataarray(
            maybe_rasterio_urlpath, engine="rasterio", chunks=chunks
        )
    return arr


def make_azimuth_time(
    product_first_line_utc_time: str,
    product_last_line_utc_time: str,
    number_of_lines: int,
) -> npt.NDArray[np.datetime64]:
    azimuth_time = pd.date_range(
        start=product_first_line_utc_time,
        end=product_last_line_utc_time,
        periods=number_of_lines,
    )
    # return a read-write array from the Pandas read-only `.values`
    return np.array(azimuth_time.values, dtype="datetime64[ns]")


def open_pol_dataset(
    measurement: esa_safe.PathOrFileType,
    annotation: esa_safe.PathOrFileType,
    fs: fsspec.AbstractFileSystem | None = None,
    attrs: dict[str, Any] = {},
    gcp: xr.Dataset | None = None,
    rasterio_chunks: dict[str, int] | None = None,
) -> xr.Dataset:
    product_information = esa_safe.parse_tag(annotation, "//productInformation")
    image_information = esa_safe.parse_tag(annotation, "//imageInformation")
    swath_timing = esa_safe.parse_tag(annotation, "//swathTiming")

    number_of_samples = image_information["numberOfSamples"]
    range_sampling_rate = product_information["rangeSamplingRate"]

    product_first_line_utc_time = image_information["productFirstLineUtcTime"]
    product_last_line_utc_time = image_information["productLastLineUtcTime"]
    number_of_lines = image_information["numberOfLines"]
    azimuth_time_interval = image_information["azimuthTimeInterval"]
    number_of_bursts = swath_timing["burstList"]["@count"]
    image_slant_range_time = image_information["slantRangeTime"]
    range_pixel_spacing = image_information["rangePixelSpacing"]

    attrs = attrs.copy()
    attrs.update(
        {
            "radar_frequency": product_information["radarFrequency"] / 10**9,
            "ascending_node_time": image_information["ascendingNodeTime"],
            "azimuth_pixel_spacing": image_information["azimuthPixelSpacing"],
            "range_pixel_spacing": range_pixel_spacing,
            "product_first_line_utc_time": product_first_line_utc_time,
            "product_last_line_utc_time": product_last_line_utc_time,
            "azimuth_time_interval": azimuth_time_interval,
            "image_slant_range_time": image_slant_range_time,
            "range_sampling_rate": range_sampling_rate,
            "incidence_angle_mid_swath": image_information["incidenceAngleMidSwath"],
        }
    )
    swap_dims = {}

    # open COG with chunks if dask is present
    try:
        import dask  # noqa

        if rasterio_chunks is None:
            rasterio_chunks = {}
    except ModuleNotFoundError:
        pass

    azimuth_time = make_azimuth_time(
        product_first_line_utc_time,
        product_last_line_utc_time,
        number_of_lines,
    )
    if number_of_bursts == 0:
        swap_dims = {"line": "azimuth_time", "pixel": "slant_range_time"}
    else:
        if "burstId" in swath_timing["burstList"]["burst"][0]:
            burst_ids = []
            for burst in swath_timing["burstList"]["burst"]:
                burst_ids.append(burst["burstId"]["$"])
            attrs["burst_ids"] = burst_ids
        lines_per_burst = swath_timing["linesPerBurst"]
        attrs.update(
            {
                "azimuth_steering_rate": product_information["azimuthSteeringRate"],
                "number_of_bursts": number_of_bursts,
                "lines_per_burst": lines_per_burst,
            }
        )
        for burst_index, burst in enumerate(swath_timing["burstList"]["burst"]):
            first_azimuth_time_burst = burst["azimuthTime"]
            azimuth_time_burst = pd.date_range(
                start=first_azimuth_time_burst,
                periods=lines_per_burst,
                freq=pd.Timedelta(azimuth_time_interval * 10**9, unit="ns"),
            )
            azimuth_time[
                lines_per_burst * burst_index : lines_per_burst * (burst_index + 1)
            ] = azimuth_time_burst

    coords = {
        "pixel": np.arange(0, number_of_samples, dtype=int),
        "line": np.arange(0, number_of_lines, dtype=int),
        # set "units" explicitly as CF conventions don't support "nanoseconds".
        # See: https://github.com/pydata/xarray/issues/4183#issuecomment-685200043
        "azimuth_time": (
            "line",
            azimuth_time,
            {},
            {"units": f"microseconds since {azimuth_time[0]}"},
        ),
    }

    if product_information["projection"] == "Slant Range":
        slant_range_time = np.linspace(
            image_slant_range_time,
            image_slant_range_time + (number_of_samples - 1) / range_sampling_rate,
            number_of_samples,
        )
        coords["slant_range_time"] = (
            "pixel",
            slant_range_time,
            {},
            {"_FillValue": None},
        )
    elif product_information["projection"] == "Ground Range":
        ground_range = np.linspace(
            0,
            range_pixel_spacing * (number_of_samples - 1),
            number_of_samples,
        )
        coords["ground_range"] = ("pixel", ground_range, {}, {"_FillValue": None})
        swap_dims = {"line": "azimuth_time", "pixel": "ground_range"}
    else:
        raise ValueError(f"unknown projection {product_information['projection']}")

    arr = open_rasterio_dataarray(measurement, fs, rasterio_chunks)

    preferred_chunks = arr.encoding["preferred_chunks"]
    # clear the encoding as many GeoTIFF details are incompatible with the CF conventions
    arr.encoding.clear()

    arr = arr.squeeze("band").drop_vars(["band", "spatial_ref"])
    arr = arr.rename({"y": "line", "x": "pixel"})
    arr = arr.assign_coords(coords)
    arr = arr.swap_dims(swap_dims)

    # setting the preferred_chunks for the output to the current arr chunks
    if rasterio_chunks is None:
        rasterio_chunks = {}
    preferred_chunks["line"] = rasterio_chunks.get("y", preferred_chunks["y"])
    preferred_chunks["pixel"] = rasterio_chunks.get("x", preferred_chunks["x"])
    for from_name, to_name in swap_dims.items():
        preferred_chunks[to_name] = preferred_chunks[from_name]

    arr.encoding.update({"preferred_chunks": preferred_chunks})

    if gcp:
        attrs.update(gcp.attrs)

    arr.attrs.update(attrs)

    return xr.Dataset(attrs=attrs, data_vars={"measurement": arr})


def find_bursts_index(
    pol_dataset: DataArrayOrDataset,
    azimuth_anx_seconds: float,
    use_center: bool = False,
) -> int:
    lines_per_burst = pol_dataset.attrs["lines_per_burst"]
    anx_datetime = np.datetime64(pol_dataset.attrs["ascending_node_time"], "ns")
    azimuth_anx_time = pd.Timedelta(int(azimuth_anx_seconds * 10**9), unit="ns")
    if use_center:
        azimuth_anx_time_center = (
            pol_dataset.azimuth_time[lines_per_burst // 2 :: lines_per_burst]
            - anx_datetime
        )
        distance = abs(azimuth_anx_time_center - azimuth_anx_time)
    else:
        azimuth_anx_time_first_line = (
            pol_dataset.azimuth_time[::lines_per_burst] - anx_datetime
        )
        distance = abs(azimuth_anx_time_first_line - azimuth_anx_time)
    return np.argmin(distance.data).item()


def crop_burst_dataset(
    pol_dataset: DataArrayOrDataset,
    burst_index: int | None = None,
    azimuth_anx_time: float | None = None,
    use_center: bool = False,
    burst_id: int | None = None,
    gcp: xr.Dataset | None = None,
) -> DataArrayOrDataset:
    """Return the measurement dataset cropped to the selected burst.

    Only one keyword between 'burst_index' and 'azimuth_anx_time' and 'burst_id' must be defined.

    :param xr.Dataset pol_dataset: measurement dataset
    :param int burst_index: burst index can take values from 1 to the number of bursts
    :param float azimuth_anx_time: azimuth anx time of first line of the bursts
    To use the center instead of the first line, set `use_center=True`
    :param bool use_center: If `True`, it uses the azimuth anx time as a reference for
    the burst center instead of the first line
    :param int burst_id: for product processed with Sentinel-1 IPF version 3.40 or higher,
    the burst can be selected using the relative burst id.
    """
    burst_definitions = (
        (burst_index is not None)
        + (azimuth_anx_time is not None)
        + (burst_id is not None)
    )
    if burst_definitions > 1:
        raise TypeError(
            "only one keyword between 'burst_index' and 'azimuth_anx_time' and 'burst_id' must be defined"
        )

    if burst_index is None:
        if azimuth_anx_time is not None:
            burst_index = find_bursts_index(
                pol_dataset, azimuth_anx_time, use_center=use_center
            )
        elif burst_id is not None:
            burst_ids = pol_dataset.attrs.get("burst_ids")
            if burst_ids is None:
                raise TypeError(
                    "'burst_ids' list can't be found in product attributes, "
                    "probably Sentinel-1 IPF processor version is older than 3.40"
                )
            try:
                burst_index = burst_ids.index(burst_id)
            except ValueError:
                raise KeyError(f"{burst_id=} not found in product {burst_ids=}")
        else:
            raise TypeError(
                "one keyword between 'burst_index' and 'azimuth_anx_time' must be defined"
            )

    if burst_index < 0 or burst_index >= pol_dataset.attrs["number_of_bursts"]:
        raise IndexError(f"{burst_index=} out of bounds")

    lines_per_burst = pol_dataset.attrs["lines_per_burst"]
    ds = pol_dataset.sel(
        line=slice(
            lines_per_burst * burst_index, lines_per_burst * (burst_index + 1) - 1
        )
    )

    ds = ds.swap_dims({"line": "azimuth_time", "pixel": "slant_range_time"})

    anx_datetime = np.datetime64(pol_dataset.attrs["ascending_node_time"], "ns")
    burst_azimuth_anx_time = ds.azimuth_time.values[0] - anx_datetime
    ds.attrs["azimuth_anx_time"] = burst_azimuth_anx_time / ONE_SECOND
    ds.attrs["burst_index"] = burst_index
    ds.attrs.pop("subgroups", None)

    if gcp is not None:
        footprint = get_footprint_linestring(ds.azimuth_time, ds.slant_range_time, gcp)
        geospatial_attrs = make_geospatial_attributes(footprint)
        ds.attrs.update(geospatial_attrs)

    if "burst_ids" in ds.attrs:
        ds.attrs["burst_id"] = ds.attrs["burst_ids"][burst_index]
        _ = ds.attrs.pop("burst_ids")

    return ds


def mosaic_slc_iw(
    slc_iw_image: DataArrayOrDataset, crop: int = 90
) -> DataArrayOrDataset:
    bursts = []
    for i in range(slc_iw_image.attrs["number_of_bursts"]):
        burst = crop_burst_dataset(slc_iw_image, burst_index=i)
        bursts.append(burst.isel(azimuth_time=slice(crop, -crop)))
    return xr.concat(bursts, dim="azimuth_time")


def calibrate_amplitude(
    digital_number: xr.DataArray,
    calibration_lut: xr.DataArray,
    **kwargs: Any,
) -> xr.DataArray:
    """Return the calibrated amplitude using the calibration LUT in the product metadata.

    :param digital_number: digital numbers to be calibrated
    :param calibration_lut: calibration LUT (sigmaNought, betaNought or gamma).

    The LUT can be opened using the measurement sub-group `calibration`
    """
    calibration_lut_mean = calibration_lut.mean()
    if np.allclose(calibration_lut_mean, calibration_lut, **kwargs):
        calibration: xr.DataArray = calibration_lut_mean.astype(np.float32)
    else:
        calibration = calibration_lut.interp(
            line=digital_number.line,
            pixel=digital_number.pixel,
        ).astype(np.float32)
        if digital_number.chunks is not None:
            calibration = calibration.chunk(digital_number.chunksizes)
    amplitude = digital_number / calibration
    amplitude.attrs.update(digital_number.attrs)
    try:
        lut_name = calibration_lut.attrs["long_name"].partition("calibration LUT")[0]
        amplitude.attrs["long_name"] = f"amplitude for {lut_name}".strip()
        amplitude.attrs["units"] = calibration.attrs["units"]
    except KeyError:
        pass
    return amplitude


def calibrate_intensity(
    digital_number: xr.DataArray,
    calibration_lut: xr.DataArray,
    as_db: bool = False,
    min_db: float | None = -40.0,
    **kwargs: Any,
) -> xr.DataArray:
    """Return the calibrated intensity using the calibration LUT in the product metadata.

    :param digital_number: digital numbers to be calibrated
    :param calibration_lut: calibration LUT (sigmaNought, betaNought or gamma).
    The LUT can be opened using the measurement sub-group `calibration`.
    :param as_db: if True, returns the data in db
    :param min_db: minimal value in db, to avoid infinity values.
    """
    amplitude = calibrate_amplitude(digital_number, calibration_lut, **kwargs)
    intensity = abs(amplitude) ** 2
    if as_db:
        intensity = 10.0 * np.log10(intensity)  # type: ignore
        if min_db is not None:
            intensity = np.maximum(intensity, min_db)  # type: ignore
        intensity.attrs.update(amplitude.attrs)
        intensity.attrs["units"] = "dB"
    else:
        intensity.attrs.update(amplitude.attrs)
        intensity.attrs["units"] = "m2 m-2"
    try:
        lut_name = amplitude.attrs["long_name"].partition("amplitude for ")[2]
        intensity.attrs["long_name"] = lut_name
    except KeyError:
        pass
    return intensity


def slant_range_time_to_ground_range(
    azimuth_time: xr.DataArray,
    slant_range_time: xr.DataArray,
    coordinate_conversion: xr.Dataset,
) -> xr.DataArray:
    """Convert slant range time to ground range using the coordinate conversion metadata.

    :param azimuth_time: azimuth time coordinates
    :param slant_range_time: slant range time
    :param coordinate_conversion: coordinate conversion dataset.
    The coordinate conversion dataset can be opened using the measurement sub-groub `coordinate_conversion`
    """
    slant_range = SPEED_OF_LIGHT / 2.0 * slant_range_time
    sr0 = coordinate_conversion.sr0.interp(azimuth_time=azimuth_time)
    srgrCoefficients = coordinate_conversion.srgrCoefficients.interp(
        azimuth_time=azimuth_time,
    )
    x = slant_range - sr0
    ground_range = (srgrCoefficients * x**srgrCoefficients.degree).sum("degree")
    return ground_range  # type: ignore


def ground_range_to_slant_range_time(
    azimuth_time: xr.DataArray,
    ground_range: xr.DataArray,
    coordinate_conversion: xr.Dataset,
) -> xr.DataArray:
    """Convert ground range to slant range time using the coordinate conversion metadata.

    :param azimuth_time: azimuth time coordinates
    :param ground_range: slant range time
    :param coordinate_conversion: coordinate conversion dataset.
    The coordinate conversion dataset can be opened using the measurement sub-groub `coordinate_conversion`
    """
    assert (coordinate_conversion.gr0 == 0.0).all()
    grsrCoefficients = coordinate_conversion.grsrCoefficients.interp(
        azimuth_time=azimuth_time,
    )
    x = ground_range
    slant_range = (grsrCoefficients * x**grsrCoefficients.degree).sum("degree")
    return 2 / SPEED_OF_LIGHT * slant_range  # type: ignore


METADATA_OPENERS = {
    "orbit": open_orbit_dataset,
    "attitude": open_attitude_dataset,
    "azimuth_fm_rate": open_azimuth_fm_rate_dataset,
    "dc_estimate": open_dc_estimate_dataset,
    "gcp": open_gcp_dataset,
    "coordinate_conversion": open_coordinate_conversion_dataset,
    "calibration": open_calibration_dataset,
    "noise_range": open_noise_range_dataset,
    "noise_azimuth": open_noise_azimuth_dataset,
    "replica": open_replica_dataset,
    "reference_replica": open_reference_replica_dataset,
    "antenna_pattern": open_antenna_pattern,
}


def do_override_product_files(
    template: str, product_files: dict[str, tuple[str, str, str, str, str]]
) -> dict[str, tuple[str, str, str, str, str]]:
    overridden_product_files = {}
    for path, description in product_files.items():
        type, prefix, swath, polarization, date = description
        ext = os.path.splitext(path)[1]
        dirname = os.path.dirname(path)
        overridden_path = template.format(**locals())
        overridden_product_files[overridden_path] = description
    return overridden_product_files


def open_sentinel1_dataset(
    product_urlpath: esa_safe.PathType,
    *,
    drop_variables: tuple[str] | None = None,
    group: str | None = None,
    fs: fsspec.AbstractFileSystem | None = None,
    storage_options: dict[str, Any] | None = None,
    check_files_exist: bool = False,
    override_product_files: str | None = None,
    parse_geospatial_attrs: bool = True,
    parse_eopf_metadata: bool = False,
    rasterio_chunks: dict[str, int] | None = None,
) -> xr.Dataset:
    if drop_variables is not None:
        warnings.warn("'drop_variables' is currently ignored")

    fs, manifest_path = get_fs_path(product_urlpath, fs, storage_options)
    product_path = os.path.dirname(manifest_path)

    with fs.open(manifest_path) as file:
        common_attrs, product_files = esa_safe.parse_manifest_sentinel1(file)

    if override_product_files:
        product_files = do_override_product_files(override_product_files, product_files)

    groups = find_available_groups(
        product_files,
        product_path,
        common_attrs["product_type"],
        check_files_exist=check_files_exist,
        fs=fs,
    )

    group, burst_index = normalise_group(group)
    absgroup = f"/{group}"
    if group != "" and group not in groups:
        raise ValueError(
            f"Invalid group {group!r}, please select one of the following groups:"
            f"\n{list(groups.keys())}"
        )

    metadata = ""

    ds = xr.Dataset(attrs=common_attrs)
    gcp = None
    if group == "":
        subgroups = list(groups)
    else:
        subgroups = [
            g[len(group) + 1 :] for g in groups if g.startswith(group) and g != group
        ]

        if group.count("/") == 1:
            with fs.open(groups[group][1]) as annotation:
                if parse_geospatial_attrs:
                    gcp = open_gcp_dataset(annotation, attrs=common_attrs)

                ds = open_pol_dataset(
                    groups[group][0],
                    annotation,
                    fs=fs,
                    attrs=common_attrs,
                    gcp=gcp,
                    rasterio_chunks=rasterio_chunks,
                )
                if parse_eopf_metadata:
                    ds.attrs["other_metadata"] = eopf_metadata.build_other_metadata(
                        annotation
                    )
                    ds.attrs["stac_discovery"] = make_sentinel1_stac_item(
                        "noid", manifest_path, annotation
                    )
        elif group.count("/") == 2:
            _, _, metadata = group.split("/", 2)
            with fs.open(groups[group][0]) as file:
                ds = METADATA_OPENERS[metadata](file, attrs=common_attrs)

    ds.attrs["group"] = absgroup
    if len(subgroups):
        ds.attrs["subgroups"] = subgroups

    if group.count("/") == 1 and burst_index is not None:
        ds = crop_burst_dataset(ds, burst_index=burst_index, gcp=gcp)

    conventions.update_attributes(ds, group=metadata)

    return ds


def make_sentinel1_stac_item(
    item_id: str,
    manifest_path: esa_safe.PathOrFileType,
    annotation: esa_safe.PathOrFileType,
    namespaces: dict[str, str] = esa_safe.SENTINEL1_NAMESPACES,
) -> dict[str, Any]:
    manifest = ElementTree.parse(manifest_path).getroot()

    product_information = esa_safe.parse_tag(annotation, "//productInformation")
    image_information = esa_safe.parse_tag(annotation, "//imageInformation")

    coordinates = [
        [float(v) for v in token.split(",")]
        for token in esa_safe.findtext(manifest, ".//gml:coordinates").split()
    ]
    coordinates += coordinates[:1]
    product_timeliness = esa_safe.findtext(
        manifest, ".//s1sarl1:productTimelinessCategory"
    )
    product_timeliness_map = {
        "Fast-24h": {
            "product:timeliness_category": "STC",
            "product:timeliness": "PT24H",
        },
        "NRT-3h": {
            "product:timeliness_category": "NRT",
            "product:timeliness": "PT3H",
        },
    }

    stac_item = {
        "type": "Feature",
        "stac_version": "1.1.0",
        "stac_extensions": [
            "https://stac-extensions.github.io/product/v0.1.0/schema.json",
            "https://stac-extensions.github.io/processing/v1.2.0/schema.json",
            "https://stac-extensions.github.io/sat/v1.0.0/schema.json",
            "https://stac-extensions.github.io/view/v1.0.0/schema.json",
            "https://stac-extensions.github.io/sar/v1.2.0/schema.json",
            "https://cs-si.github.io/eopf-stac-extension/v1.2.0/schema.json",
        ],
        "id": item_id,
        "properties": {
            "datetime": None,
            "start_datetime": esa_safe.findtext(manifest, ".//safe:startTime") + "Z",
            "end_datetime": esa_safe.findtext(manifest, ".//safe:stopTime") + "Z",
            "created": manifest.find(
                ".//safe:processing", namespaces=namespaces
            ).attrib["stop"]  # type: ignore
            + "Z",
            "platform": "sentinel-1"
            + esa_safe.findtext(manifest, ".//safe:platform/safe:number").lower(),
            "instruments": ["sar"],
            "constellation": "sentinel-1",
            "product:type": esa_safe.findtext(manifest, ".//s1sarl1:productType"),
            "product:timeliness_category": product_timeliness_map[product_timeliness][
                "product:timeliness_category"
            ],
            "product:timeliness": product_timeliness_map[product_timeliness][
                "product:timeliness"
            ],
            "processing:software": manifest.find(
                ".//safe:software", namespaces=namespaces
            ).attrib,  # type: ignore
            "sat:platform_international_designator": esa_safe.findtext(
                manifest, ".//safe:nssdcIdentifier"
            ),
            "sat:absolute_orbit": int(
                esa_safe.findall(manifest, ".//safe:orbitNumber")[0]
            ),
            "sat:relative_orbit": int(
                esa_safe.findall(manifest, ".//safe:relativeOrbitNumber")[0]
            ),
            "sat:orbit_state": esa_safe.findtext(manifest, ".//s1:pass").lower(),
            "sat:anx_datetime": esa_safe.findtext(manifest, ".//s1:ascendingNodeTime")
            + "Z",
            "view:incidence_angle": image_information["incidenceAngleMidSwath"],
            "sar:polarizations": esa_safe.findall(
                manifest, ".//s1sarl1:transmitterReceiverPolarisation"
            ),
            "sar:instrument_mode": esa_safe.findtext(
                manifest, ".//s1sarl1:instrumentMode/s1sarl1:mode"
            ),
            "sar:frequency_band": "C",
            "sar:center_frequency": product_information["radarFrequency"] / 1e9,
            "sar:pixel_spacing_range": image_information["rangePixelSpacing"],
            "sar:pixel_spacing_azimuth": image_information["azimuthPixelSpacing"],
            "sar:observation_direction": "right",
            "sar:beam_ids": esa_safe.findall(
                manifest, ".//s1sarl1:instrumentMode/s1sarl1:swath"
            ),
            "eopf:datatake_id": esa_safe.findtext(
                manifest, ".//s1sarl1:missionDataTakeID"
            ),
        },
        "geometry": {"type": "Polygon", "coordinates": [coordinates]},
        "links": [],
        "assets": {},
    }
    return stac_item



import pathlib
from typing import Any

import pytest

from xarray_sentinel import reformat

DATA_FOLDER = pathlib.Path(__file__).parent / "data"


def test_to_group_zarr(tmpdir: Any) -> None:
    product_path = (
        DATA_FOLDER
        / "S1B_IW_SLC__1SDV_20210401T052622_20210401T052650_026269_032297_EFA4.SAFE"
    )
    tmp_path = str(tmpdir.join("tmp.zarr"))

    reformat.to_group_zarr(product_path, tmp_path)


@pytest.mark.xfail
def test_to_group_netcdf(tmpdir: Any) -> None:
    product_path = (
        DATA_FOLDER
        / "S1B_IW_SLC__1SDV_20210401T052622_20210401T052650_026269_032297_EFA4.SAFE"
    )
    tmp_path = str(tmpdir.join("tmp.nc"))

    reformat.to_group_netcdf(product_path, tmp_path)



from typing import Any, Dict

import xarray as xr

from . import esa_safe


def to_group_zarr(
    product_path: esa_safe.PathType,
    output_store: Any,
    groups: Dict[str, str] | None = None,
    **kwargs: Any,
) -> None:
    root = xr.open_dataset(product_path, engine="sentinel-1")
    root.to_zarr(output_store, mode="w", **kwargs)

    if groups is None:
        groups = {g: g for g in root.attrs["subgroups"]}

    for group_out, group_in in groups.items():
        try:
            group_ds = xr.open_dataset(
                product_path, engine="sentinel-1", group=group_in
            )
            group_ds.to_zarr(output_store, mode="a", group=group_out, **kwargs)
        except FileNotFoundError:
            pass


# Apparently there is no way to save SLC images because "netcdf4" doesn't support complex data
# and "h5netcdf" crashes on an issue related to dimension names
def to_group_netcdf(
    product_path: esa_safe.PathType,
    output_store: str,
    groups: Dict[str, str] | None = None,
    **kwargs: Any,
) -> None:
    root = xr.open_dataset(product_path, engine="sentinel-1")
    root.to_netcdf(output_store, mode="w", **kwargs)

    if groups is None:
        groups = {g: g for g in root.attrs["subgroups"]}

    for group_out, group_in in groups.items():
        try:
            group_ds = xr.open_dataset(
                product_path, engine="sentinel-1", group=group_in
            )
            group_ds.to_netcdf(output_store, mode="a", group=group_out, **kwargs)
        except FileNotFoundError:
            pass



import os
from typing import Any

import fsspec
import xarray as xr

from . import sentinel1


class Sentinel1Backend(xr.backends.common.BackendEntrypoint):
    def open_dataset(  # type: ignore
        self,
        filename_or_obj: str,
        drop_variables: tuple[str] | None = None,
        group: str | None = None,
        storage_options: dict[str, Any] | None = None,
        override_product_files: str | None = None,
        fs: fsspec.AbstractFileSystem | None = None,
        check_files_exist: bool = False,
        parse_geospatial_attrs: bool = True,
        rasterio_chunks: dict[str, int] | None = None,
    ) -> xr.Dataset:
        ds = sentinel1.open_sentinel1_dataset(
            filename_or_obj,
            drop_variables=drop_variables,
            group=group,
            storage_options=storage_options,
            override_product_files=override_product_files,
            fs=fs,
            check_files_exist=check_files_exist,
            parse_geospatial_attrs=parse_geospatial_attrs,
            rasterio_chunks=rasterio_chunks,
        )
        return ds

    def guess_can_open(self, filename_or_obj: Any) -> bool:
        try:
            _, ext = os.path.splitext(filename_or_obj)
        except TypeError:
            return False
        return ext.lower() in {".safe", ".safe/"}



import pathlib
from typing import Any

import fsspec
import xarray as xr

from . import esa_safe, sentinel1


def open_datatree(
    product_urlpath: esa_safe.PathType,
    *,
    fs: fsspec.AbstractFileSystem | None = None,
    storage_options: dict[str, Any] | None = None,
    check_files_exist: bool = False,
    override_product_files: str | None = None,
    **kwargs: Any,
) -> xr.DataTree:
    product_name = pathlib.Path(product_urlpath).stem
    root = sentinel1.open_sentinel1_dataset(
        product_urlpath,
        fs=fs,
        storage_options=storage_options,
        check_files_exist=check_files_exist,
        override_product_files=override_product_files,
    )
    xarray_sentinel_groups = root.attrs["subgroups"]
    dt = xr.DataTree()
    for xarray_sentinel_group in xarray_sentinel_groups:
        swath, _, pol_group = xarray_sentinel_group.partition("/")
        pol, _, dataset = pol_group.partition("/")
        eopf_product_name = f"{product_name}_{swath}_{pol.upper()}"
        if not pol:
            continue
        if not dataset:
            measurement_ds = sentinel1.open_sentinel1_dataset(
                product_urlpath,
                fs=fs,
                storage_options=storage_options,
                check_files_exist=check_files_exist,
                override_product_files=override_product_files,
                group=xarray_sentinel_group,
                parse_eopf_metadata=True,
            ).rename(measurement="slc")
            if eopf_product_name not in dt.children:
                product_ds = xr.Dataset(
                    attrs={
                        "other_metadata": measurement_ds.attrs["other_metadata"],
                        "stac_discovery": measurement_ds.attrs["stac_discovery"],
                    }
                )
                dt[f"{eopf_product_name}"] = product_ds
            measurement_ds.attrs.clear()
            dt[f"{eopf_product_name}/measurements"] = measurement_ds
        elif dataset in {"orbit", "attitude", "dc_estimate", "gcp"}:
            ds = sentinel1.open_sentinel1_dataset(
                product_urlpath,
                fs=fs,
                storage_options=storage_options,
                check_files_exist=check_files_exist,
                group=xarray_sentinel_group,
                override_product_files=override_product_files,
            )
            if dataset == "dc_estimate":
                dataset = "doppler_centroid"
            ds.attrs.clear()
            dt[f"{eopf_product_name}/conditions/{dataset}"] = ds
        elif dataset in {"calibration", "noise_range", "noise_azimuth"}:
            ds = sentinel1.open_sentinel1_dataset(
                product_urlpath,
                fs=fs,
                storage_options=storage_options,
                check_files_exist=check_files_exist,
                group=xarray_sentinel_group,
                override_product_files=override_product_files,
            )
            ds.attrs.clear()
            dt[f"{eopf_product_name}/quality/{dataset}"] = ds
        else:
            print(f"Skipping {xarray_sentinel_group=}")

    return dt



import re
import warnings
from typing import Any

import fsspec
import numpy as np

from . import esa_safe


# taken from pydantic.alias_generators, MIT license
def to_snake(camel: str) -> str:
    """Convert a PascalCase, camelCase, or kebab-case string to snake_case.

    Args:
        camel: The string to convert.

    Returns
    -------
        The converted string in snake_case.
    """
    # Handle the sequence of uppercase letters followed by a lowercase letter
    snake = re.sub(
        r"([A-Z]+)([A-Z][a-z])", lambda m: f"{m.group(1)}_{m.group(2)}", camel
    )
    # Insert an underscore between a lowercase letter and an uppercase letter
    snake = re.sub(r"([a-z])([A-Z])", lambda m: f"{m.group(1)}_{m.group(2)}", snake)
    # Insert an underscore between a digit and an uppercase letter
    snake = re.sub(r"([0-9])([A-Z])", lambda m: f"{m.group(1)}_{m.group(2)}", snake)
    # Insert an underscore between a lowercase letter and a digit
    snake = re.sub(r"([a-z])([0-9])", lambda m: f"{m.group(1)}_{m.group(2)}", snake)
    # Replace hyphens with underscores to handle kebab-case
    snake = snake.replace("-", "_")
    return snake.lower()


def to_snake_recursive(
    struct: dict[str, Any] | list[Any],
) -> dict[str, Any] | list[Any]:
    if isinstance(struct, dict):
        struct = {to_snake(k): to_snake_recursive(v) for k, v in struct.items()}
    elif isinstance(struct, list):
        struct = [to_snake_recursive(v) for v in struct]
    return struct


def fix_lists(struct: Any) -> Any:
    fixed: Any
    if isinstance(struct, dict):
        fixed = {}
        for k, v in struct.items():
            if k == "@count":
                continue
            if k[-5:] == "_list":
                try:
                    fixed[k] = fix_lists(struct[k][k[:-5]])
                except Exception:
                    fixed[k] = fix_lists(fix_lists(struct[k]))
            else:
                fixed[k] = fix_lists(struct[k])
    elif isinstance(struct, list):
        fixed = [fix_lists(v) for v in struct]
    else:
        fixed = struct
    return fixed


def filter_metadata_dict(image_information: dict[str, Any]) -> dict[str, Any]:
    image_information = to_snake_recursive(image_information)  # type: ignore
    image_information = fix_lists(image_information)
    return image_information


def build_azimuth_fm_rate_list(
    azimuth_fm_rate_list: list[dict[str, Any] | list[Any]],
) -> list[dict[str, Any] | list[Any]]:
    azimuth_fm_rate_list_out: list[dict[str, Any] | list[Any]] = []
    for item in azimuth_fm_rate_list:
        azimuth_fm_rate_polynomial_str = item["azimuth_fm_rate_polynomial"]["$"]  # type: ignore
        azimuth_fm_rate_list_out.append(
            {
                "azimuth_time": item["azimuth_time"],  # type: ignore
                "azimuth_fm_rate_polynomial": np.fromstring(
                    azimuth_fm_rate_polynomial_str, sep=" "
                ).tolist(),
            }
        )
    return azimuth_fm_rate_list_out


def build_general_annotation(general_annotation: dict[str, Any]) -> dict[str, Any]:
    general_annotation = filter_metadata_dict(general_annotation)
    _ = general_annotation.pop("orbit_list")
    _ = general_annotation.pop("attitude_list")
    general_annotation["azimuth_fm_rate_list"] = build_azimuth_fm_rate_list(
        general_annotation["azimuth_fm_rate_list"]
    )
    return general_annotation


def build_other_metadata(annotation_urlpath: esa_safe.PathType) -> dict[str, Any]:
    warnings.warn("This is an unofficial, alpha converter", UserWarning)
    with fsspec.open(annotation_urlpath) as fp:
        quality_information = esa_safe.parse_tag(fp, "//qualityInformation")
        quality_information = filter_metadata_dict(quality_information)
        general_annotation = esa_safe.parse_tag(fp, "//generalAnnotation")
        general_annotation = build_general_annotation(general_annotation)
        image_information = esa_safe.parse_tag(fp, "//imageAnnotation")
        image_information = filter_metadata_dict(image_information)
        swath_merginig = esa_safe.parse_tag(fp, "//swathMerging")
        swath_merginig = filter_metadata_dict(swath_merginig)
        swath_timing = esa_safe.parse_tag(fp, "//swathTiming")
        swath_timing = filter_metadata_dict(swath_timing)

    other_metadata = {
        "quality_information": quality_information,
        "general_annotation": general_annotation,
        "image_annotation": image_information,
        "swath_timing": swath_timing,
        "swath_merginig": swath_merginig,
    }
    return other_metadata



import pathlib

import fsspec
import pytest
import xarray as xr

from xarray_sentinel import sentinel1

DATA_FOLDER = pathlib.Path(__file__).parent / "data"

SLC_IW = (
    DATA_FOLDER
    / "S1B_IW_SLC__1SDV_20210401T052622_20210401T052650_026269_032297_EFA4.SAFE"
)
SLC_IW1_VV_annotation = (
    DATA_FOLDER
    / "S1B_IW_SLC__1SDV_20210401T052622_20210401T052650_026269_032297_EFA4.SAFE"
    / "annotation"
    / "s1b-iw1-slc-vv-20210401t052624-20210401t052649-026269-032297-004.xml"
)
SLC_IW1_VV_calibration = (
    DATA_FOLDER
    / "S1B_IW_SLC__1SDV_20210401T052622_20210401T052650_026269_032297_EFA4.SAFE"
    / "annotation"
    / "calibration"
    / "calibration-s1b-iw1-slc-vv-20210401t052624-20210401t052649-026269-032297-004.xml"
)
SLC_IW1_VV_measurement = (
    DATA_FOLDER
    / "S1B_IW_SLC__1SDV_20210401T052622_20210401T052650_026269_032297_EFA4.SAFE"
    / "measurement"
    / "s1b-iw1-slc-vv-20210401t052624-20210401t052649-026269-032297-004.tiff"
)
SLC_S3_VV_annotation = (
    DATA_FOLDER
    / "S1A_S3_SLC__1SDV_20210401T152855_20210401T152914_037258_04638E_6001.SAFE"
    / "annotation"
    / "s1a-s3-slc-vv-20210401t152855-20210401t152914-037258-04638e-002.xml"
)
SLC_S3_VV_measurement = (
    DATA_FOLDER
    / "S1A_S3_SLC__1SDV_20210401T152855_20210401T152914_037258_04638E_6001.SAFE"
    / "measurement"
    / "s1a-s3-slc-vv-20210401t152855-20210401t152914-037258-04638e-002.tiff"
)
GRD_IW_VV_annotation = (
    DATA_FOLDER
    / "S1B_IW_GRDH_1SDV_20210401T052623_20210401T052648_026269_032297_ECC8.SAFE"
    / "annotation"
    / "s1b-iw-grd-vv-20210401t052623-20210401t052648-026269-032297-001.xml"
)


def test_get_fs_path() -> None:
    fs, path = sentinel1.get_fs_path(SLC_IW)

    assert isinstance(fs, fsspec.AbstractFileSystem)
    assert path == str(SLC_IW / "manifest.safe")

    fs2, path2 = sentinel1.get_fs_path(path, fs=fs)

    assert fs2 is fs
    assert path2 is path

    with pytest.raises(ValueError):
        sentinel1.get_fs_path("dummy*")

    with pytest.raises(ValueError):
        sentinel1.get_fs_path("*")


def test_open_dataset_zip_metadata() -> None:
    zip_path = (
        DATA_FOLDER
        / "S1B_IW_SLC__1SDV_20210401T052622_20210401T052650_026269_032297_EFA4.zip"
    )
    zip_urlpath = f"zip://*/manifest.safe::{zip_path}"
    expected_groups = {
        "IW1",
        "IW1/VV",
        "IW1/VV/gcp",
        "IW1/VV/attitude",
        "IW1/VV/orbit",
        "IW1/VV/calibration",
    }

    res = sentinel1.open_sentinel1_dataset(zip_urlpath)

    assert isinstance(res, xr.Dataset)
    assert set(res.attrs["subgroups"]) >= expected_groups

    res = sentinel1.open_sentinel1_dataset(zip_urlpath, group="IW1/VV/orbit")

    assert isinstance(res, xr.Dataset)
    assert res.sizes == {"axis": 3, "azimuth_time": 17}


@pytest.mark.xfail
def test_open_dataset_zip_data() -> None:
    zip_path = (
        DATA_FOLDER
        / "S1B_IW_SLC__1SDV_20210401T052622_20210401T052650_026269_032297_EFA4.zip"
    )
    zip_urlpath = f"zip://*/manifest.safe::{zip_path}"

    res = sentinel1.open_sentinel1_dataset(zip_urlpath, group="IW1/VV/0")

    assert isinstance(res, xr.Dataset)
    assert res.sizes == {"slant_range_time": 21632, "azimuth_time": 1501}
    assert abs(res.measurement[:40, :40]).mean() >= 0



import pathlib
from typing import Any, Dict

import pytest
import xarray as xr
from cfchecker import cfchecks

pytest.importorskip("netCDF4")

DATA_FOLDER = pathlib.Path(__file__).parent / "data"


def cfcheck(path: str) -> Dict[str, int]:
    (
        badc,
        coards,
        debug,
        uploader,
        useFileName,
        regionnames,
        standardName,
        areaTypes,
        cacheDir,
        cacheTables,
        cacheTime,
        version,
        files,
    ) = cfchecks.getargs(["cfchecks", path])

    inst = cfchecks.CFChecker(
        uploader=uploader,
        useFileName=useFileName,
        badc=badc,
        coards=coards,
        cfRegionNamesXML=regionnames,
        cfStandardNamesXML=standardName,
        cfAreaTypesXML=areaTypes,
        cacheDir=cacheDir,
        cacheTables=cacheTables,
        cacheTime=cacheTime,
        version=version,
        debug=debug,
    )
    for file in files:
        try:
            inst.checker(file)
        except cfchecks.FatalCheckerError:
            print("Checking of file %s aborted due to error" % file)

    totals: Dict[str, int] = inst.get_total_counts()

    return totals


def test_cfcheck_grd(tmpdir: Any) -> None:
    product_path = (
        DATA_FOLDER
        / "S1B_IW_GRDH_1SDV_20210401T052623_20210401T052648_026269_032297_ECC8.SAFE"
    )

    groups = [""]
    while groups:
        group = groups.pop()
        try:
            ds = xr.open_dataset(product_path, engine="sentinel-1", group=group)
            groups.extend(f"{group}/{g}" for g in ds.attrs.get("subgroups", []))
        except FileNotFoundError:
            continue
        nc_path = tmpdir.join(group.replace("/", "-") + ".nc")
        ds.to_netcdf(nc_path)

        totals = cfcheck(str(nc_path))

        assert totals["FATAL"] + totals["ERROR"] + totals["WARN"] == 0



import pathlib

from xarray_sentinel import eopf_metadata

DATA_FOLDER = pathlib.Path(__file__).parent / "data"

SLC_S3 = (
    DATA_FOLDER
    / "S1A_S3_SLC__1SDV_20210401T152855_20210401T152914_037258_04638E_6001.SAFE"
)
SLC_S3_VH_annotation = (
    SLC_S3
    / "annotation"
    / "s1a-s3-slc-vh-20210401t152855-20210401t152914-037258-04638e-001.xml"
)


def test_to_snake_recursive() -> None:
    metadata = {"qualityInformation": {"qualityDataList": [{"qualityData": 2}]}}
    expected = {"quality_information": {"quality_data_list": [{"quality_data": 2}]}}

    res = eopf_metadata.to_snake_recursive(metadata)

    assert res == expected


def test_build_other_metadata() -> None:
    res = eopf_metadata.build_other_metadata(SLC_S3_VH_annotation)
    expected_quality_information = {
        "product_quality_index": 0.0,
        "quality_data_list": [
            {
                "azimuth_time": "2021-04-01T15:28:55.111501",
                "downlink_quality": {
                    "i_input_data_mean": 0.06302300840616226,
                    "q_input_data_mean": 0.114902101457119,
                    "input_data_mean_outside_nominal_range_flag": False,
                    "i_input_data_std_dev": 1.674363970756531,
                    "q_input_data_std_dev": 1.515843987464905,
                    "input_data_st_dev_outside_nominal_range_flag": False,
                    "num_downlink_input_data_gaps": 0,
                    "downlink_gaps_in_input_data_significant_flag": False,
                    "num_downlink_input_missing_lines": 0,
                    "downlink_missing_lines_significant_flag": False,
                    "num_instrument_input_data_gaps": 0,
                    "instrument_gaps_in_input_data_significant_flag": False,
                    "num_instrument_input_missing_lines": 0,
                    "instrument_missing_lines_significant_flag": False,
                    "num_ssb_error_input_data_gaps": 0,
                    "ssb_error_gaps_in_input_data_significant_flag": False,
                    "num_ssb_error_input_missing_lines": 0,
                    "ssb_error_missing_lines_significant_flag": False,
                    "chirp_source_used": "Nominal",
                    "pg_source_used": "Extracted",
                    "rrf_spectrum_used": "Extended Tapered",
                    "replica_reconstruction_failed_flag": False,
                    "mean_pg_product_amplitude": 0.8810305595397949,
                    "std_dev_pg_product_amplitude": 0.006521929986774921,
                    "mean_pg_product_phase": 0.07730674743652344,
                    "std_dev_pg_product_phase": 0.01562887243926525,
                    "pg_product_derivation_failed_flag": False,
                    "invalid_downlink_params_flag": True,
                },
                "raw_data_analysis_quality": {
                    "i_bias": 0.06302300840616226,
                    "i_bias_significance_flag": True,
                    "q_bias": 0.114902101457119,
                    "q_bias_significance_flag": True,
                    "iq_gain_imbalance": 1.104575037956238,
                    "iq_gain_significance_flag": True,
                    "iq_quadrature_departure": -0.5162643194198608,
                    "iq_quadrature_departure_significance_flag": True,
                },
                "doppler_centroid_quality": {
                    "dc_method": "Data Analysis",
                    "doppler_centroid_uncertain_flag": False,
                },
                "image_quality": {
                    "image_statistics": {
                        "output_data_mean": {"re": -0.02062067, "im": 0.01445807},
                        "output_data_std_dev": {"re": 6.993358, "im": 6.996273},
                    },
                    "output_data_mean_outside_nominal_range_flag": True,
                    "output_data_st_dev_outside_nominal_range_flag": True,
                },
            }
        ],
    }

    assert len(res) == 5

    assert res["quality_information"] == expected_quality_information



# Copyright 2021-2026, B-Open Solutions srl and the xarray-sentinel authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    # NOTE: the `version.py` file must not be present in the git repository
    #   as it is generated by setuptools at install time
    from .version import __version__
except ImportError:  # pragma: no cover
    # Local copy or not installed with setuptools
    __version__ = "999"

from .esa_safe import make_stac_item
from .sentinel1 import (
    calibrate_amplitude,
    calibrate_intensity,
    crop_burst_dataset,
    get_footprint_linestring,
    ground_range_to_slant_range_time,
    mosaic_slc_iw,
    open_sentinel1_dataset,
    slant_range_time_to_ground_range,
)

__all__ = [
    "__version__",
    "calibrate_amplitude",
    "calibrate_intensity",
    "crop_burst_dataset",
    "get_footprint_linestring",
    "ground_range_to_slant_range_time",
    "make_stac_item",
    "mosaic_slc_iw",
    "open_sentinel1_dataset",
    "slant_range_time_to_ground_range",
]
